import requests
import json
import os
import time
from typing import List, Optional
from requests.exceptions import RequestException

from webscout.AIbase import ImageProvider

class NexraImager(ImageProvider):
    """Image provider for Nexra API"""

    AVAILABLE_MODELS = {
        "standard": ["emi", "stablediffusion-1.5", "stablediffusion-2.1", "sdxl-lora", "dalle", "dalle2", "dalle-mini", "flux", "midjourney"],
        "prodia": [
            "dreamshaperXL10_alpha2.safetensors [c8afe2ef]",
            "dynavisionXL_0411.safetensors [c39cc051]",
            "juggernautXL_v45.safetensors [e75f5471]",
            "realismEngineSDXL_v10.safetensors [af771c3f]",
            "sd_xl_base_1.0.safetensors [be9edd61]",
            "animagineXLV3_v30.safetensors [75f2f05b]",
            "sd_xl_base_1.0_inpainting_0.1.safetensors [5679a81a]",
            "turbovisionXL_v431.safetensors [78890989]",
            "devlishphotorealism_sdxl15.safetensors [77cba69f]",
            "realvisxlV40.safetensors [f7fdcb51]"
        ]
    }

    def __init__(self, timeout: int = 60, proxies: dict = {}):
        self.url = "https://nexra.aryahcr.cc/api/image/complements"
        self.headers = {"Content-Type": "application/json"}
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        self.session.proxies.update(proxies)
        self.timeout = timeout
        self.prompt: str = "AI-generated image - webscout"
        self.image_extension: str = "png"

    def generate(
        self, prompt: str, model: str = "flux", amount: int = 1,
        max_retries: int = 3, retry_delay: int = 5,
        additional_params: Optional[dict] = None
    ) -> List[bytes]:
        assert bool(prompt), "Prompt cannot be null"
        assert isinstance(amount, int) and amount > 0, "Amount should be a positive integer"

        self.prompt = prompt
        response = []

        payload = {
            "prompt": prompt,
            "model": "prodia" if model in self.AVAILABLE_MODELS["prodia"] else model,
        }

        if model in self.AVAILABLE_MODELS["prodia"]:
            payload["data"] = {
                "model": model,
                "steps": 25,
                "cfg_scale": 7,
                "sampler": "DPM++ 2M Karras",
                "negative_prompt": ""
            }
        if additional_params:
            payload.update(additional_params)

        for _ in range(max_retries):
            try:
                resp = self.session.post(self.url, json=payload, timeout=self.timeout)
                resp.raise_for_status()

                # Remove leading underscores and then parse JSON
                response_data = json.loads(resp.text.lstrip("_"))

                if response_data.get("status") and "images" in response_data:
                    for image_url in response_data["images"]:
                        img_resp = requests.get(image_url)
                        img_resp.raise_for_status()
                        response.append(img_resp.content)
                    break
                else:
                    raise Exception("Failed to generate image: " + str(response_data))
            except json.JSONDecodeError as json_err:
                print(f"JSON Decode Error: {json_err}")
                print(f"Raw response: {resp.text}")
                if _ == max_retries - 1:
                    raise
            except RequestException as e:
                print(f"Request Exception: {e}")
                if _ == max_retries - 1:
                    raise
            print(f"Retrying in {retry_delay} seconds...")
            time.sleep(retry_delay)

        return response

    def save(
        self,
        response: List[bytes],
        name: str = None,
        dir: str = os.getcwd(),
        filenames_prefix: str = "",
    ) -> List[str]:
        assert isinstance(response, list), f"Response should be a list, not {type(response)}"
        name = self.prompt if name is None else name

        filenames = []
        for i, image in enumerate(response):
            filename = f"{filenames_prefix}{name}_{i}.{self.image_extension}"
            filepath = os.path.join(dir, filename)

            with open(filepath, "wb") as fh:
                fh.write(image)
            filenames.append(filename)

        return filenames

if __name__ == "__main__":
    bot = NexraImager()
    resp_standard = bot.generate("AI-generated image - webscout", "midjourney", 1)
    print(bot.save(resp_standard))



from datetime import date
import json
# import logging
import time
from typing import Any, Dict, Optional
import requests
from webscout import WEBS, Julius

class FunctionCallingAgent:
    def __init__(self, 
                 tools: list = None):
        self.ai = Julius(timeout=300, intro=None)
        self.tools = tools if tools is not None else []
        self.knowledge_cutoff = "September 2022" 

    def function_call_handler(self, message_text: str) -> dict:
        system_message = self._generate_system_message(message_text)
        response = self.ai.chat(system_message)
        # logging.info(f"Raw response: {response}")
        return self._parse_function_call(response)

    def _generate_system_message(self, user_message: str) -> str:
        tools_description = ""
        for tool in self.tools:
            tools_description += f"- {tool['function']['name']}: {tool['function'].get('description', '')}\n"
            tools_description += "    Parameters:\n"
            for key, value in tool['function']['parameters']['properties'].items():
                tools_description += f"      - {key}: {value.get('description', '')} ({value.get('type')})\n"
        
        current_date = date.today().strftime("%B %d, %Y")
        return f"""Today is {current_date}. Your knowledge is current up to {self.knowledge_cutoff}.

You are an advanced AI assistant tasked with analyzing user requests and determining the most appropriate action. You have access to the following tools:

{tools_description}

Instructions:
1. Carefully analyze the user's request.
2. If the user request is about controlling the web browser (opening websites, searching, playing videos, etc.), ALWAYS use the 'browser_control' tool.
3. If the user request is something else, determine if any of the other provided tools are necessary to fulfill the request.
4. If a tool is needed, select the MOST APPROPRIATE one. Do not use a tool if it's not directly relevant to the user's request.
5. If you decide to use a tool, respond ONLY with a JSON object in this format:
   {{
     "tool_name": "name_of_the_tool",
     "tool_input": {{
       "param1": "value1",
       "param2": "value2"
     }}
   }}

   - Use the exact tool name as listed above.
   - Include only the necessary parameters for the chosen tool.
   - Do not include any explanations or additional text outside the JSON object.

6. If no tool is needed and you can answer directly, respond with:
   {{
     "tool_name": "general_ai",
     "tool_input": {{
       "question": "user's original question"
     }}
   }}

User Request: {user_message}

Your Response (JSON only):"""

    def _parse_function_call(self, response: str) -> dict:
        try:
            # Ensure to locate the start and end of the JSON structure
            start_idx = response.find("{")
            end_idx = response.rfind("}") + 1

            if start_idx == -1 or end_idx == -1:
                raise ValueError("No valid JSON structure found in the response.")

            # Extract the JSON string
            response_json_str = response[start_idx:end_idx]

            # Load the JSON string
            parsed_response = json.loads(response_json_str)

            # Check for the expected format
            if "tool_name" in parsed_response and "tool_input" in parsed_response:
                return parsed_response

            # If not in the expected format, try to convert
            for key, value in parsed_response.items():
                if isinstance(value, dict):
                    return {
                        "tool_name": key,
                        "tool_input": value
                    }

            # If the response structure is still incorrect
            raise ValueError("Invalid response structure: missing required fields.")

        except (ValueError, json.JSONDecodeError) as e:
            # logging.error(f"Error parsing function call: {e}")
            return {"error": str(e)}


    def execute_function(self, function_call_data: dict) -> str:
        function_name = function_call_data.get("tool_name")
        arguments = function_call_data.get("tool_input", {})

        if not isinstance(arguments, dict):
            # logging.error("Invalid arguments format.")
            return "Invalid arguments format."

        # logging.info(f"Executing function: {function_name} with arguments: {arguments}")


# Example usage
if __name__ == "__main__":
    # Configure logging
    # logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

    tools = [
        {
            "type": "function",
            "function": {
                "name": "web_search",
                "description": "Search the web for current information on a given query",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "The search query to be executed"
                        }
                    },
                    "required": ["query"]
                }
            }
        },
        {
            "type": "function",
            "function": {
                "name": "get_user_detail",
                "description": "Get the user's name and age.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "name": {
                            "type": "string",
                            "description": "The user's name."
                        },
                        "age": {
                            "type": "integer",
                            "description": "The user's age."
                        }
                    },
                    "required": ["name", "age"]
                }
            }
        },
        {
            "type": "function",
            "function": {
                "name": "general_ai",
                "description": "Use AI to answer general questions or perform tasks not requiring external tools",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "question": {
                            "type": "string",
                            "description": "The question or task for the AI to process"
                        }
                    },
                    "required": ["question"]
                }
            }
        }
    ]

    agent = FunctionCallingAgent(tools=tools)
    
    # Test cases
    test_messages = [
        "What's the weather like in New York today?",
        "Who won the last FIFA World Cup?",
        "Can you explain quantum computing?",
        "What are the latest developments in AI?",
        "Tell me a joke about programming.",
        "What's the meaning of life?",
        "Get user details name as John and age as 30"
    ]

    for message in test_messages:
        print(f"\nProcessing: {message}")
        function_call_data = agent.function_call_handler(message)
        print(f"Function Call Data: {function_call_data}")

        if "error" not in function_call_data:
            result = agent.execute_function(function_call_data)
            print(f"Function Execution Result: {result}")
        else:
            print(f"Error: {function_call_data['error']}")


import requests
import re
import json
import os
from typing import List, Dict
from webscout.AIutel import Optimizers, Conversation, AwesomePrompts
from webscout.AIbase import Provider
from webscout import exceptions
from rich import print

class ChatGPTES(Provider):
    """
    A class to interact with the ChatGPT.es API.
    """

    SUPPORTED_MODELS = {
        'gpt-4o', 'gpt-4o-mini', 'chatgpt-4o-latest'
    }

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        model: str = "chatgpt-4o-latest",  # Default model
        system_prompt: str = "You are a helpful assistant.",
    ):
        """
        Initializes the ChatGPT.es API with given parameters.
        """
        if model not in self.SUPPORTED_MODELS:
            raise ValueError(f"Unsupported model: {model}. Choose from: {self.SUPPORTED_MODELS}")

        self.session = requests.Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = 'https://chatgpt.es/wp-admin/admin-ajax.php'
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.system_prompt = system_prompt
        self.model = model
        self.initial_headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) '
                          'AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36',
            'Referer': 'https://www.google.com/',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,'
                      'image/avif,image/webp,image/apng,*/*;q=0.8,'
                      'application/signed-exchange;v=b3;q=0.7',
        }
        self.post_headers = {
            'User-Agent': self.initial_headers['User-Agent'],
            'Referer': 'https://chatgpt.es/',
            'Origin': 'https://chatgpt.es',
            'Accept': '*/*',
            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
        }
        self.nonce = None
        self.post_id = None

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )

        # Conversation setup
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def get_nonce_and_post_id(self):
        """
        Retrieves the nonce and post ID from the ChatGPT.es website.
        """
        try:
            response = self.session.get('https://chatgpt.es/', headers=self.initial_headers, timeout=self.timeout)
            response.raise_for_status()
        except requests.RequestException as e:
            raise ConnectionError(f"Failed to retrieve nonce and post_id: {e}")

        nonce_match = re.search(r'data-nonce="(.+?)"', response.text)
        post_id_match = re.search(r'data-post-id="(.+?)"', response.text)

        if not nonce_match or not post_id_match:
            raise ValueError("Failed to parse nonce or post_id from the response.")

        self.nonce = nonce_match.group(1)
        self.post_id = post_id_match.group(1)

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> dict:
        """
        Chat with ChatGPT.es

        Args:
            prompt (str): Prompt to be sent.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            raw (bool, optional): Stream back raw response as received. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.

        Returns:
            dict: Response dictionary.
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                optimizer_func = getattr(Optimizers, optimizer)
                conversation_prompt = optimizer_func(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise ValueError(f"Optimizer '{optimizer}' is not supported. "
                                 f"Available optimizers: {list(self.__available_optimizers)}")

        # Retrieve nonce and post_id if they are not set
        if not self.nonce or not self.post_id:
            self.get_nonce_and_post_id()

        messages = [
            {"role": "user", "content": conversation_prompt},
        ]

        # Prepare conversation history
        conversation = ["Human: strictly respond in the same language as my prompt, preferably English"]
        for msg in messages:
            role = "Human" if msg['role'] == "user" else "AI"
            conversation.append(f"{role}: {msg['content']}")

        payload = {
            '_wpnonce': self.nonce,
            'post_id': self.post_id,
            'url': 'https://chatgpt.es',
            'action': 'wpaicg_chat_shortcode_message',
            'message': messages[-1]['content'],
            'bot_id': '0',
            'chatbot_identity': 'shortcode',
            'wpaicg_chat_client_id': os.urandom(5).hex(),
            'wpaicg_chat_history': json.dumps(conversation)
        }

        try:
            response = self.session.post(
                self.api_endpoint,
                headers=self.post_headers,
                data=payload,
                timeout=self.timeout
            )
            response.raise_for_status()
        except requests.RequestException as e:
            raise ConnectionError(f"Failed to send request to ChatGPT.es: {e}")

        try:
            response_data = response.json()
        except json.JSONDecodeError:
            raise ValueError(f"Invalid JSON response: {response.text}")

        if not isinstance(response_data, dict):
            raise TypeError(f"Expected response_data to be a dict, got {type(response_data)}")

        # Extract the message directly from the 'data' key
        message = response_data.get('data')
        if not isinstance(message, str):
            raise KeyError("Missing 'data' key in response or 'data' is not a string")

        self.last_response.update(dict(text=message))
        self.conversation.update_chat_history(
            prompt, self.get_message(self.last_response)
        )
        return self.last_response

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        """
        Generate response as a string.

        Args:
            prompt (str): Prompt to be sent.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.

        Returns:
            str: Response generated.
        """
        response = self.ask(
            prompt,
            stream=stream,
            optimizer=optimizer,
            conversationally=conversationally,
        )
        return self.get_message(response)

    def get_message(self, response: dict) -> str:
        """
        Retrieves message only from response.

        Args:
            response (dict): Response generated by `self.ask`.

        Returns:
            str: Message extracted.
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response["text"]

if __name__ == "__main__":
    ai = ChatGPTES() 
    response = ai.chat(input(">>> "))
    for chunk in response:
        print(chunk, end="", flush=True)




import requests
import json
import uuid
import os
from typing import List

from webscout.AIbase import ImageProvider

class AmigoImager(ImageProvider):
    """
    Image provider for AmigoChat.io.
    """
    AVAILABLE_MODELS = ["dalle-e-3", "flux-pro", "flux-realism"] 
    def __init__(self, timeout: int = 60, proxies: dict = {}):
        """Initializes the AmigoImager class.

        Args:
            timeout (int, optional): HTTP request timeout in seconds. Defaults to 60.
            proxies (dict, optional): HTTP request proxies. Defaults to {}.
        """
        self.url = "https://api.amigochat.io/v1/images/generations"
        self.headers = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br, zstd",
            "Accept-Language": "en-US,en;q=0.9,en-IN;q=0.8",
            "Authorization": "Bearer ",  # Empty
            "Content-Type": "application/json; charset=utf-8",
            "DNT": "1",
            "Origin": "https://amigochat.io",
            "Referer": "https://amigochat.io/",
            "Sec-CH-UA": '"Microsoft Edge";v="129", "Not=A?Brand";v="8", "Chromium";v="129"',
            "Sec-CH-UA-Mobile": "?0",
            "Sec-CH-UA-Platform": '"Windows"',
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-site",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36 Edg/129.0.0.0",
            "X-Device-Language": "en-US",
            "X-Device-Platform": "web",
            "X-Device-UUID": str(uuid.uuid4()),
            "X-Device-Version": "1.0.22"
        }
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        self.session.proxies.update(proxies)
        self.timeout = timeout
        self.prompt: str = "AI-generated image - webscout"
        self.image_extension: str = "png"

    def generate(self, prompt: str, amount: int = 1, model: str = "flux-pro") -> List[str]:
        """Generate image from prompt

        Args:
            prompt (str): Image description.
            amount (int, optional): Total images to be generated. Defaults to 1.
            model (str, optional): Model to use for generating images. Defaults to "flux-pro".

        Returns:
            List[str]: List of generated image URLs.
        """
        assert bool(prompt), "Prompt cannot be null"
        assert isinstance(amount, int), f"Amount should be an integer only not {type(amount)}"
        assert amount > 0, "Amount should be greater than 0"
        assert model in self.AVAILABLE_MODELS, f"Model should be one of {self.AVAILABLE_MODELS}"

        self.prompt = prompt
        response = []

        for _ in range(amount):
            # JSON payload for the request
            payload = {
                "prompt": prompt,
                "model": model,
                "personaId": "image-generator"
            }
            
            try:
                # Sending the POST request
                resp = requests.post(self.url, headers=self.headers, data=json.dumps(payload), timeout=self.timeout)
                resp.raise_for_status()

                # Process the response
                response_data = resp.json()
                image_url = response_data['data'][0]['url']
                response.append(image_url)

            except requests.exceptions.RequestException as e:
                print(f"An error occurred: {e}")
                raise

        return response
    
    def save(
        self,
        response: List[str],  # List of image URLs
        name: str = None,
        dir: str = os.getcwd(),
        filenames_prefix: str = "",
    ) -> List[str]:
        """Save generated images

        Args:
            response (List[str]): List of generated image URLs.
            name (str): Filename for the images. Defaults to the last prompt.
            dir (str, optional): Directory for saving images. Defaults to os.getcwd().
            filenames_prefix (str, optional): String to be prefixed at each filename to be returned.

        Returns:
            List[str]: List of saved filenames.
        """
        assert isinstance(response, list), f"Response should be of {list} not {type(response)}"
        name = self.prompt if name is None else name

        filenames = []
        count = 0
        for img_url in response:
            def complete_path():
                count_value = "" if count == 0 else f"_{count}"
                return os.path.join(dir, name + count_value + "." + self.image_extension)

            while os.path.isfile(complete_path()):
                count += 1

            absolute_path_to_file = complete_path()
            filenames.append(filenames_prefix + os.path.split(absolute_path_to_file)[1])

            # Download and save the image
            try:
                img_response = requests.get(img_url, stream=True, timeout=self.timeout)
                img_response.raise_for_status()

                with open(absolute_path_to_file, "wb") as fh:
                    for chunk in img_response.iter_content(chunk_size=8192):
                        fh.write(chunk)
            except requests.exceptions.RequestException as e:
                print(f"An error occurred while downloading image from {img_url}: {e}")
                raise

        return filenames

# Example usage
if __name__ == "__main__":
    bot = AmigoImager()
    try:
        resp = bot.generate("A shiny red sports car speeding down a scenic mountain road", 1)
        print(bot.save(resp))
    except Exception as e:
        print(f"An error occurred: {e}")


from abc import ABC
from abc import abstractmethod
from typing import AsyncGenerator, List, Union, Generator

class Provider(ABC):
    """Base class for text-based AI providers"""

    @abstractmethod
    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> dict:
        """Chat with AI

        Args:
            prompt (str): Prompt to be sent
            stream (bool, optional): Flag for streaming response. Defaults to False.
            raw (bool, optional): Stream back raw response as received
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
           dict : {}
        ```json
        {
            "completion": "\nNext: domestic cat breeds with short hair >>",
            "stop_reason": null,
            "truncated": false,
            "stop": null,
            "model": "llama-2-13b-chat",
            "log_id": "cmpl-3kYiYxSNDvgMShSzFooz6t",
            "exception": null
        }
        ```
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

    @abstractmethod
    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be sent
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
            str: Response generated
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

    @abstractmethod
    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

class AsyncProvider(ABC):
    """Asynchronous base class for text-based AI providers"""

    @abstractmethod
    async def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> dict:
        """Asynchronously chat with AI

        Args:
            prompt (str): Prompt to be sent
            stream (bool, optional): Flag for streaming response. Defaults to False.
            raw (bool, optional): Stream back raw response as received
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
           dict : {}
        ```json
        {
            "completion": "\nNext: domestic cat breeds with short hair >>",
            "stop_reason": null,
            "truncated": false,
            "stop": null,
            "model": "llama-2-13b-chat",
            "log_id": "cmpl-3kYiYxSNDvgMShSzFooz6t",
            "exception": null
        }
        ```
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

    @abstractmethod
    async def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        """Asynchronously generate response `str`
        Args:
            prompt (str): Prompt to be sent
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
            str: Response generated
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

    @abstractmethod
    async def get_message(self, response: dict) -> str:
        """Asynchronously retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

class TTSProvider(ABC):
    """Base class for text-to-speech providers"""

    @abstractmethod
    def tts(self, text: str) -> Union[bytes, str, Generator[bytes, None, None]]:
        """
        Converts text to speech.

        Args:
            text (str): The text to be converted to speech.

        Returns:
            Union[bytes, str, Generator[bytes, None, None]]: 
                - The raw audio bytes if the provider supports returning audio directly.
                - The filename of the saved audio file if the provider saves the audio to a file.
                - A generator yielding chunks of audio bytes if the provider supports streaming audio.
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

class AsyncTTSProvider(ABC):
    """Asynchronous base class for text-to-speech providers"""

    @abstractmethod
    async def tts(self, text: str) -> Union[bytes, str, AsyncGenerator[bytes, None]]:
        """
        Asynchronously converts text to speech.

        Args:
            text (str): The text to be converted to speech.

        Returns:
            Union[bytes, str, AsyncGenerator[bytes, None]]: 
                - The raw audio bytes if the provider supports returning audio directly.
                - The filename of the saved audio file if the provider saves the audio to a file.
                - An asynchronous generator yielding chunks of audio bytes if the provider supports streaming audio.
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

class ImageProvider(ABC):
    """Base class for text-to-image providers"""

    @abstractmethod
    def generate(self, prompt: str, amount: int = 1) -> List[bytes]:
        """Generates images from a text prompt.

        Args:
            prompt (str): The text prompt describing the desired image.
            amount (int, optional): The number of images to generate. Defaults to 1.

        Returns:
            List[bytes]: A list of generated images in bytes format.
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

    @abstractmethod
    def save(self, response: List[bytes], name: str = None, dir: str = None) -> List[str]:
        """Saves the generated images to files.

        Args:
            response (List[bytes]): The list of generated images in bytes format.
            name (str, optional): The base filename for the images. Defaults to None.
            dir (str, optional): The directory to save the images. Defaults to None.

        Returns:
            List[str]: A list of the saved filenames.
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

class AsyncImageProvider(ABC):
    """Base class for asynchronous text-to-image providers"""

    @abstractmethod
    async def generate(self, prompt: str, amount: int = 1) -> Union[AsyncGenerator[bytes, None], List[bytes]]:
        """Asynchronously generates images from a text prompt.

        Args:
            prompt (str): The text prompt describing the desired image.
            amount (int, optional): The number of images to generate. Defaults to 1.

        Returns:
            Union[AsyncGenerator[bytes, None], List[bytes]]: 
                - An asynchronous generator yielding image bytes.
                - A list of image bytes if not streaming.
        """
        raise NotImplementedError("Method needs to be implemented in subclass")

    @abstractmethod
    async def save(self, response: Union[AsyncGenerator[bytes, None], List[bytes]], 
                   name: str = None, dir: str = None) -> List[str]:
        """Asynchronously saves the generated images to files.

        Args:
            response (Union[AsyncGenerator[bytes, None], List[bytes]]): 
                - The asynchronous generator yielding images in bytes format (if streaming).
                - The list of generated images in bytes format (if not streaming).
            name (str, optional): The base filename for the images. Defaults to None.
            dir (str, optional): The directory to save the images. Defaults to None.

        Returns:
            List[str]: A list of the saved filenames.
        """
        raise NotImplementedError("Method needs to be implemented in subclass")


import warnings
from datetime import time
import os
import sys
import subprocess
import logging
import psutil
from huggingface_hub import hf_hub_url, cached_download
import colorlog
import ollama  # Import ollama for interactive chat
import argparse  # Import argparse for command-line arguments

# Suppress specific warnings
warnings.filterwarnings(
    "ignore", category=FutureWarning, module="huggingface_hub.file_download"
)

# Configure logging with colors
handler = colorlog.StreamHandler()
handler.setFormatter(
    colorlog.ColoredFormatter(
        "%(log_color)s%(asctime)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
    )
)

logger = colorlog.getLogger(__name__)
if not logger.hasHandlers():
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

# Redirect warnings to the logger but avoid duplication
logging.captureWarnings(True)
py_warnings_logger = logging.getLogger("py.warnings")
if not py_warnings_logger.hasHandlers():
    py_warnings_logger.addHandler(handler)


def show_art():
    logger.info("Made with love in India")


def usage():
    logger.info("Usage: python script.py -m <model_path> -g <gguf_file>")
    logger.info("Options:")
    logger.info("  -m <model_path>    Set the path to the model")
    logger.info("  -g <gguf_file>     Set the GGUF file name")
    logger.info("  -h                 Display this help and exit")


def is_model_downloaded(logging_name, download_log):
    if not os.path.exists(download_log):
        return False
    with open(download_log, "r") as f:
        for line in f:
            if line.strip() == logging_name:
                return True
    return False


def log_downloaded_model(logging_name, download_log):
    with open(download_log, "a") as f:
        f.write(logging_name + "\n")


def is_model_created(model_name):
    result = subprocess.run(["ollama", "list"], stdout=subprocess.PIPE)
    return model_name in result.stdout.decode("utf-8")


def download_model(repo_id, filename, token, cache_dir="downloads"):
    url = hf_hub_url(repo_id, filename)
    filepath = cached_download(
        url, cache_dir=cache_dir, force_filename=filename, use_auth_token=token
    )
    return filepath


def is_ollama_running():
    for proc in psutil.process_iter(["name"]):
        if proc.info["name"] in ["ollama", "ollama.exe"]:
            return True
    return False


def main(model_path=None, gguf_file=None):  # Modified to handle both CLI and non-CLI
    show_art()

    # Parse command-line arguments if provided
    parser = argparse.ArgumentParser(description="Download and create an Ollama model")
    parser.add_argument("-m", "--model_path", help="Path to the model on Hugging Face Hub")
    parser.add_argument("-g", "--gguf_file", help="Name of the GGUF file")
    args = parser.parse_args()

    # Use arguments from command line or function parameters
    model_path = args.model_path if args.model_path else model_path
    gguf_file = args.gguf_file if args.gguf_file else gguf_file

    if not model_path or not gguf_file:
        logger.error("Error: model_path and gguf_file are required.")
        usage()
        sys.exit(2)

    model_name = gguf_file.split(".Q4")[0]
    download_log = "downloaded_models.log"
    logging_name = f"{model_path}_{model_name}"

    # Ensure the log file exists
    if not os.path.exists(download_log):
        with open(download_log, 'w') as f:
            pass

    # Check if huggingface-hub is installed, and install it if not
    try:
        subprocess.check_output(['pip', 'show', 'huggingface-hub'])
    except subprocess.CalledProcessError:
        logger.info("Installing huggingface-hub...")
        subprocess.check_call(['pip', 'install', '-U', 'huggingface_hub[cli]'])
    else:
        logger.info("huggingface-hub is already installed.")

    # Check if the model has already been downloaded
    if is_model_downloaded(logging_name, download_log):
        logger.info(f"Model {logging_name} has already been downloaded. Skipping download.")
    else:
        logger.info(f"Downloading model {logging_name}...")
        token = os.getenv('HUGGINGFACE_TOKEN', None)
        if not token:
            logger.warning("Warning: HUGGINGFACE_TOKEN environment variable is not set. Using None.")
            token = None

        filepath = download_model(model_path, gguf_file, token)
        log_downloaded_model(logging_name, download_log)
        logger.info(f"Model {logging_name} downloaded and logged.")

    # Check if Ollama is installed, and install it if not
    try:
        subprocess.check_output(['ollama', '--version'])
    except subprocess.CalledProcessError:
        logger.info("Installing Ollama...")
        subprocess.check_call(['curl', '-fsSL', 'https://ollama.com/install.sh', '|', 'sh'])
    else:
        logger.info("Ollama is already installed.")

    # Check if Ollama is already running
    if is_ollama_running():
        logger.info("Ollama is already running. Skipping the start.")
    else:
        logger.info("Starting Ollama...")
        subprocess.Popen(['ollama', 'serve'])

        while not is_ollama_running():
            logger.info("Waiting for Ollama to start...")
            time.sleep(1)

        logger.info("Ollama has started.")

    # Check if the model has already been created
    if is_model_created(model_name):
        logger.info(f"Model {model_name} is already created. Skipping creation.")
    else:
        logger.info(f"Creating model {model_name}...")
        with open('Modelfile', 'w') as f:
            f.write(f"FROM ./downloads/{gguf_file}")
        subprocess.check_call(['ollama', 'create', model_name, '-f', 'Modelfile'])
        logger.info(f"Model {model_name} created.")

    logger.info(f"model name is > {model_name}")
    logger.info(f"Use Ollama run {model_name}")

if __name__ == "__main__":
    main()


# webscout/providers/__init__.py
from .PI import *
from .Llama import LLAMA
from .Cohere import Cohere
from .Reka import REKA
from .Groq import GROQ
from .Groq import AsyncGROQ
from .Openai import OPENAI
from .Openai import AsyncOPENAI
from .Koboldai import KOBOLDAI
from .Koboldai import AsyncKOBOLDAI
from .Perplexity import *
from .Blackboxai import BLACKBOXAI 
from .Blackboxai import AsyncBLACKBOXAI
from .Phind import PhindSearch 
from .Phind import Phindv2
from .ai4chat import *
from .Gemini import GEMINI
from .BasedGPT import BasedGPT
from .Deepseek import DeepSeek
from .Deepinfra import DeepInfra
from .Farfalle import *
from .cleeai import *
from .OLLAMA import OLLAMA
from .Andi import AndiSearch
from .PizzaGPT import *
from .Llama3 import *
from .DARKAI import *
from .koala import *
from .RUBIKSAI import * 
from .meta import *
from .DiscordRocks import *
from .felo_search import *
from .xdash import *
from .julius import *
from .Youchat import *
from .yep import *
from .Cloudflare import *
from .turboseek import *
from .Free2GPT import *
from .EDITEE import *
from .TeachAnything import *
from .AI21 import *
from .Chatify import *
from .x0gpt import *
from .cerebras import *
from .lepton import *
from .geminiapi import *
from .elmo import *
from .genspark import *
from .upstage import *
from .Bing import *
from .GPTWeb import *
from .aigames import *
from .llamatutor import *
from .promptrefine import *
from .twitterclone import *
from .tutorai import *
from .bixin import *
from .ChatGPTES import *
from .Amigo import *
from .prefind import *
from .bagoodex import *
from .ChatHub import *
from .aimathgpt import *
from .gaurish import *
from .geminiprorealtime import *
from .NinjaChat import *
__all__ = [
    'Farfalle',
    'LLAMA', 
    'Cohere',
    'REKA',
    'GROQ',
    'AsyncGROQ',
    'OPENAI',
    'AsyncOPENAI',
    'KOBOLDAI',
    'AsyncKOBOLDAI',
    'Perplexity',
    'BLACKBOXAI', 
    'AsyncBLACKBOXAI',
    'PhindSearch', 
    'Felo',
    'GEMINI',
    'BasedGPT',
    'DeepSeek',
    'DeepInfra',
    'AI4Chat',
    'Phindv2',
    'OLLAMA',
    'AndiSearch',
    'PIZZAGPT',
    'LLAMA3',
    'DARKAI',
    'KOALA',
    'RUBIKSAI',
    'Meta',
    'DiscordRocks',
    'PiAI',
    'XDASH',
    'Julius',
    'YouChat',
    'YEPCHAT',
    'Cloudflare',
    'TurboSeek',
    'Editee',
    'TeachAnything',
    'AI21',
    'Chatify',
    'X0GPT',
    'Cerebras',
    'Lepton',
    'GEMINIAPI',
    'Cleeai',
    'Elmo',
    'Genspark',
    'Upstage',
    'Free2GPT',
    'Bing',
    'GPTWeb',
    'AIGameIO',
    'LlamaTutor',
    'PromptRefine',
    'AIUncensored',
    'TutorAI',
    'Bixin',
    'ChatGPTES',
    'AmigoChat',
    'PrefindAI',
    'Bagoodex',
    'ChatHub',
    'AIMathGPT',
    'GaurishCerebras',
    'GeminiPro',
    'NinjaChat',


]


import requests
import json
import os
from typing import List, Dict, Optional

from webscout.AIbase import ImageProvider
from webscout import exceptions  # Import exceptions module


class NinjaImager(ImageProvider):
    """
    Image provider for NinjaChat.ai.
    """

    AVAILABLE_MODELS = {
        "stable-diffusion": "https://www.ninjachat.ai/api/image-generator",
        "flux-dev": "https://www.ninjachat.ai/api/flux-image-generator",
    }

    def __init__(self, timeout: int = 60, proxies: dict = {}):
        """Initializes the NinjaChatImager class."""
        self.headers = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br, zstd",
            "Accept-Language": "en-US,en;q=0.9,en-IN;q=0.8",
            "Content-Type": "application/json",
            "DNT": "1",
            "Origin": "https://www.ninjachat.ai",
            "Priority": "u=1, i",
            "Referer": "https://www.ninjachat.ai/image-generation",
            "Sec-CH-UA": '"Microsoft Edge";v="129", "Not=A?Brand";v="8", "Chromium";v="129"',
            "Sec-CH-UA-Mobile": "?0",
            "Sec-CH-UA-Platform": '"Windows"',
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36 Edg/129.0.0.0",
        }
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        self.session.proxies.update(proxies)
        self.timeout = timeout
        self.prompt = "AI-generated image - webscout"
        self.image_extension = "png"  # Default extension

    def generate(self, prompt: str, amount: int = 1, model: str = "flux-dev") -> List[str]:
        """Generate images from a prompt."""

        assert bool(prompt), "Prompt cannot be null"
        assert isinstance(amount, int) and amount > 0, "Amount should be a positive integer"

        if model not in self.AVAILABLE_MODELS:
            raise exceptions.ModelNotFoundError(f"Model '{model}' not found. Available models: {', '.join(self.AVAILABLE_MODELS)}")

        self.prompt = prompt  # Store the prompt
        url = self.AVAILABLE_MODELS[model]

        payload = {
            "prompt": prompt,
            "model": model if model == "flux-dev" else "stable-diffusion", # Pass model name to API if needed
            "negativePrompt": "", #Use negative prompt from API's data structure
            "cfg": 7,
            "aspectRatio": "1:1",
            "outputFormat": self.image_extension,
            "numOutputs": amount,
            "outputQuality": 90
        }


        image_urls = []
        try:
            with requests.post(url, headers=self.headers, json=payload, timeout=self.timeout) as response:
                if response.status_code != 200:
                    raise exceptions.FailedToGenerateResponseError(f"Request failed with status code: {response.status_code}, {response.text}")  # Raise Webscout exception

                data = response.json()

                if 'output' not in data:
                    raise exceptions.InvalidResponseError("Invalid API response format: 'output' key missing.")

                for img_url in data['output']:
                    image_urls.append(img_url)

        except requests.exceptions.RequestException as e:
            raise exceptions.APIConnectionError(f"An error occurred during the request: {e}") # More specific exception
        except json.JSONDecodeError as e:
            raise exceptions.InvalidResponseError(f"Failed to parse JSON response: {e}")

        return image_urls


    def save(
        self,
        response: List[str],
        name: str = None,
        dir: str = os.getcwd(),
        filenames_prefix: str = "",
    ) -> List[str]:
        """Saves generated images."""

        assert isinstance(response, list), f"Response should be a list, not {type(response)}"
        name = self.prompt if name is None else name

        filenames = []
        count = 0
        for img_url in response:
            def complete_path():
                count_value = "" if count == 0 else f"_{count}"
                return os.path.join(dir, name + count_value + "." + self.image_extension)

            while os.path.isfile(complete_path()):
                count += 1

            absolute_path_to_file = complete_path()
            filenames.append(filenames_prefix + os.path.split(absolute_path_to_file)[1])
            try:
                with requests.get(img_url, stream=True, timeout=self.timeout) as img_response:
                    img_response.raise_for_status()
                    with open(absolute_path_to_file, "wb") as f:
                        for chunk in img_response.iter_content(chunk_size=8192):
                            f.write(chunk)

            except requests.exceptions.RequestException as e:
                raise exceptions.FailedToSaveImageError(f"An error occurred while downloading/saving image: {e}")

        return filenames



if __name__ == "__main__":
    bot = NinjaImager()
    try:
        resp = bot.generate("A shiny red sports car speeding down a scenic mountain road", 1)
        print(bot.save(resp))
    except Exception as e:
        print(f"An error occurred: {e}")


import asyncio
from types import TracebackType
from typing import Dict, List, Optional, Type, Union

from .webscout_search import WEBS


class AsyncWEBS(WEBS):
    def __init__(
        self,
        headers: Optional[Dict[str, str]] = None,
        proxy: Optional[str] = None,
        proxies: Union[Dict[str, str], str, None] = None,  # deprecated
        timeout: Optional[int] = 10,
    ) -> None:
        """Initialize the AsyncWEBS object.

        Args:
            headers (dict, optional): Dictionary of headers for the HTTP client. Defaults to None.
            proxy (str, optional): proxy for the HTTP client, supports http/https/socks5 protocols.
                example: "http://user:pass@example.com:3128". Defaults to None.
            timeout (int, optional): Timeout value for the HTTP client. Defaults to 10.
        """
        super().__init__(headers=headers, proxy=proxy, proxies=proxies, timeout=timeout)
        self._loop = asyncio.get_running_loop()
        self._executor = super()._executor

    async def __aenter__(self) -> "AsyncWEBS":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    async def achat(self, keywords: str, model: str = "gpt-3.5") -> str:
        """Initiates async chat session with webscout AI.

        Args:
            keywords (str): The initial message or question to send to the AI.
            model (str): The model to use: "gpt-3.5", "claude-3-haiku", "llama-3-70b", "mixtral-8x7b".
                Defaults to "gpt-3.5".

        Returns:
            str: The response from the AI.
        """
        result = await self._loop.run_in_executor(self._executor, super().chat, keywords, model)
        return result

    async def atext(
        self,
        keywords: str,
        region: str = "wt-wt",
        safesearch: str = "moderate",
        timelimit: Optional[str] = None,
        backend: str = "api",
        max_results: Optional[int] = None,
    ) -> List[Dict[str, str]]:
        """webscout async text search. Query params: https://duckduckgo.com/params.

        Args:
            keywords: keywords for query.
            region: wt-wt, us-en, uk-en, ru-ru, etc. Defaults to "wt-wt".
            safesearch: on, moderate, off. Defaults to "moderate".
            timelimit: d, w, m, y. Defaults to None.
            backend: api, html, lite. Defaults to api.
                api - collect data from https://duckduckgo.com,
                html - collect data from https://html.duckduckgo.com,
                lite - collect data from https://lite.duckduckgo.com.
            max_results: max number of results. If None, returns results only from the first response. Defaults to None.

        Returns:
            List of dictionaries with search results, or None if there was an error.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor, super().text, keywords, region, safesearch, timelimit, backend, max_results
        )
        return result

    async def aimages(
        self,
        keywords: str,
        region: str = "wt-wt",
        safesearch: str = "moderate",
        timelimit: Optional[str] = None,
        size: Optional[str] = None,
        color: Optional[str] = None,
        type_image: Optional[str] = None,
        layout: Optional[str] = None,
        license_image: Optional[str] = None,
        max_results: Optional[int] = None,
    ) -> List[Dict[str, str]]:
        """webscout async images search. Query params: https://duckduckgo.com/params.

        Args:
            keywords: keywords for query.
            region: wt-wt, us-en, uk-en, ru-ru, etc. Defaults to "wt-wt".
            safesearch: on, moderate, off. Defaults to "moderate".
            timelimit: Day, Week, Month, Year. Defaults to None.
            size: Small, Medium, Large, Wallpaper. Defaults to None.
            color: color, Monochrome, Red, Orange, Yellow, Green, Blue,
                Purple, Pink, Brown, Black, Gray, Teal, White. Defaults to None.
            type_image: photo, clipart, gif, transparent, line.
                Defaults to None.
            layout: Square, Tall, Wide. Defaults to None.
            license_image: any (All Creative Commons), Public (PublicDomain),
                Share (Free to Share and Use), ShareCommercially (Free to Share and Use Commercially),
                Modify (Free to Modify, Share, and Use), ModifyCommercially (Free to Modify, Share, and
                Use Commercially). Defaults to None.
            max_results: max number of results. If None, returns results only from the first response. Defaults to None.

        Returns:
            List of dictionaries with images search results.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor,
            super().images,
            keywords,
            region,
            safesearch,
            timelimit,
            size,
            color,
            type_image,
            layout,
            license_image,
            max_results,
        )
        return result

    async def avideos(
        self,
        keywords: str,
        region: str = "wt-wt",
        safesearch: str = "moderate",
        timelimit: Optional[str] = None,
        resolution: Optional[str] = None,
        duration: Optional[str] = None,
        license_videos: Optional[str] = None,
        max_results: Optional[int] = None,
    ) -> List[Dict[str, str]]:
        """webscout async videos search. Query params: https://duckduckgo.com/params.

        Args:
            keywords: keywords for query.
            region: wt-wt, us-en, uk-en, ru-ru, etc. Defaults to "wt-wt".
            safesearch: on, moderate, off. Defaults to "moderate".
            timelimit: d, w, m. Defaults to None.
            resolution: high, standart. Defaults to None.
            duration: short, medium, long. Defaults to None.
            license_videos: creativeCommon, youtube. Defaults to None.
            max_results: max number of results. If None, returns results only from the first response. Defaults to None.

        Returns:
            List of dictionaries with videos search results.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor,
            super().videos,
            keywords,
            region,
            safesearch,
            timelimit,
            resolution,
            duration,
            license_videos,
            max_results,
        )
        return result

    async def anews(
        self,
        keywords: str,
        region: str = "wt-wt",
        safesearch: str = "moderate",
        timelimit: Optional[str] = None,
        max_results: Optional[int] = None,
    ) -> List[Dict[str, str]]:
        """webscout async news search. Query params: https://duckduckgo.com/params.

        Args:
            keywords: keywords for query.
            region: wt-wt, us-en, uk-en, ru-ru, etc. Defaults to "wt-wt".
            safesearch: on, moderate, off. Defaults to "moderate".
            timelimit: d, w, m. Defaults to None.
            max_results: max number of results. If None, returns results only from the first response. Defaults to None.

        Returns:
            List of dictionaries with news search results.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor,
            super().news,
            keywords,
            region,
            safesearch,
            timelimit,
            max_results,
        )
        return result

    async def aanswers(
        self,
        keywords: str,
    ) -> List[Dict[str, str]]:
        """webscout async instant answers. Query params: https://duckduckgo.com/params.

        Args:
            keywords: keywords for query,

        Returns:
            List of dictionaries with instant answers results.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor,
            super().answers,
            keywords,
        )
        return result

    async def asuggestions(
        self,
        keywords: str,
        region: str = "wt-wt",
    ) -> List[Dict[str, str]]:
        """webscout async suggestions. Query params: https://duckduckgo.com/params.

        Args:
            keywords: keywords for query.
            region: wt-wt, us-en, uk-en, ru-ru, etc. Defaults to "wt-wt".

        Returns:
            List of dictionaries with suggestions results.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor,
            super().suggestions,
            keywords,
            region,
        )
        return result

    async def amaps(
        self,
        keywords: str,
        place: Optional[str] = None,
        street: Optional[str] = None,
        city: Optional[str] = None,
        county: Optional[str] = None,
        state: Optional[str] = None,
        country: Optional[str] = None,
        postalcode: Optional[str] = None,
        latitude: Optional[str] = None,
        longitude: Optional[str] = None,
        radius: int = 0,
        max_results: Optional[int] = None,
    ) -> List[Dict[str, str]]:
        """webscout async maps search. Query params: https://duckduckgo.com/params.

        Args:
            keywords: keywords for query
            place: if set, the other parameters are not used. Defaults to None.
            street: house number/street. Defaults to None.
            city: city of search. Defaults to None.
            county: county of search. Defaults to None.
            state: state of search. Defaults to None.
            country: country of search. Defaults to None.
            postalcode: postalcode of search. Defaults to None.
            latitude: geographic coordinate (north-south position). Defaults to None.
            longitude: geographic coordinate (east-west position); if latitude and
                longitude are set, the other parameters are not used. Defaults to None.
            radius: expand the search square by the distance in kilometers. Defaults to 0.
            max_results: max number of results. If None, returns results only from the first response. Defaults to None.

        Returns:
            List of dictionaries with maps search results, or None if there was an error.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor,
            super().maps,
            keywords,
            place,
            street,
            city,
            county,
            state,
            country,
            postalcode,
            latitude,
            longitude,
            radius,
            max_results,
        )
        return result

    async def atranslate(
        self,
        keywords: Union[List[str], str],
        from_: Optional[str] = None,
        to: str = "en",
    ) -> List[Dict[str, str]]:
        """webscout async translate.

        Args:
            keywords: string or list of strings to translate.
            from_: translate from (defaults automatically). Defaults to None.
            to: what language to translate. Defaults to "en".

        Returns:
            List od dictionaries with translated keywords.

        Raises:
            DuckDuckGoSearchException: Base exception for duckduckgo_search errors.
            RatelimitException: Inherits from DuckDuckGoSearchException, raised for exceeding API request rate limits.
            TimeoutException: Inherits from DuckDuckGoSearchException, raised for API request timeouts.
        """
        result = await self._loop.run_in_executor(
            self._executor,
            super().translate,
            keywords,
            from_,
            to,
        )
        return result


from .deepinfra import *
from .PollinationsAI import *
from .aiforce import *
from .blackboximage import *
from .Nexra import *
from .huggingface import *
from .artbit import *
from .amigo import *
from .WebSimAI import *
from .imgninza import *


import requests
import json
from typing import Any, Dict, Optional

from webscout.AIutel import Optimizers
from webscout.AIutel import Conversation
from webscout.AIutel import AwesomePrompts
from webscout.AIbase import Provider, AsyncProvider
from webscout import exceptions


class Farfalle(Provider):
    """
    A class to interact with the Farfalle.dev API.
    """

    AVAILABLE_MODELS = [
        "gpt-3.5-turbo",

    ]

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        model: str = "gpt-3.5-turbo", 
    ) -> None:
        """
        Initializes the Farfalle.dev API with given parameters.

        Args:
            is_conversation (bool, optional): Flag for chatting conversationally. Defaults to True.
            max_tokens (int, optional): Maximum number of tokens to be generated upon completion. 
                                        Defaults to 600.
            timeout (int, optional): Http request timeout. Defaults to 30.
            intro (str, optional): Conversation introductory prompt. Defaults to None.
            filepath (str, optional): Path to file containing conversation history. Defaults to None.
            update_file (bool, optional): Add new prompts and responses to the file. Defaults to True.
            proxies (dict, optional): Http request proxies. Defaults to {}.
            history_offset (int, optional): Limit conversation history to this number of last texts. 
                                            Defaults to 10250.
            act (str|int, optional): Awesome prompt key or index. (Used as intro). Defaults to None.
            model (str, optional): AI model to use. Defaults to "gpt-3.5-turbo". 
                                    Options: "gpt-3.5-turbo", "gpt-4"
        """
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.session = requests.Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = "https://api.farfalle.dev/chat"
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.model = model
        self.headers = {
            "accept": "text/event-stream",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://www.farfalle.dev",
            "priority": "u=1, i",
            "referer": "https://www.farfalle.dev/",
            "sec-ch-ua": '"Not)A;Brand";v="99", "Microsoft Edge";v="127", "Chromium";v="127"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-site",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36 Edg/127.0.0.0"
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.session.headers.update(self.headers)
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> Dict[str, Any]:
        """
        Sends a prompt to the Farfalle.dev API and returns the response.

        Args:
            prompt: The text prompt to generate text from.
            stream (bool, optional): Whether to stream the response. Defaults to False.
            raw (bool, optional): Whether to return the raw response. Defaults to False.
            optimizer (str, optional): The name of the optimizer to use. Defaults to None.
            conversationally (bool, optional): Whether to chat conversationally. Defaults to False.

        Returns:
            The response from the API.
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )
        
        payload = {
            "query": conversation_prompt,
            "model": self.model 
        }

        def for_stream():
            response = self.session.post(
                self.api_endpoint, json=payload, headers=self.headers, stream=True, timeout=self.timeout
            )

            if not response.ok:
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to generate response - ({response.status_code}, {response.reason})"
                )
            
            streaming_response = ""
            for line in response.iter_lines():
                if line:
                    decoded_line = line.decode('utf-8')
                    if decoded_line.startswith("data:"):
                        data = decoded_line[len("data:"):].strip()
                        if data:
                            try:
                                event = json.loads(data)
                                if event.get("event") == "final-response":
                                    message = event['data'].get('message', '')
                                    streaming_response += message
                                    yield message if raw else dict(text=streaming_response)
                            except json.decoder.JSONDecodeError:
                                continue
            self.last_response.update(dict(text=streaming_response))
            self.conversation.update_chat_history(
                prompt, self.get_message(self.last_response)
            )
        def for_non_stream():
            for _ in for_stream():
                pass
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
            str: Response generated
        """

        def for_stream():
            for response in self.ask(
                prompt, True, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            return self.get_message(
                self.ask(
                    prompt,
                    False,
                    optimizer=optimizer,
                    conversationally=conversationally,
                )
            )

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response["text"]
if __name__ == "__main__":
    from rich import print

    ai = Farfalle() 
    # Stream the response
    response = ai.chat("who is pm of india")
    for chunk in response:
        print(chunk, end="", flush=True)


import os
import json
from typing import Optional
import uuid
import requests
import cloudscraper

from webscout.AIutel import Optimizers
from webscout.AIutel import Conversation
from webscout.AIutel import AwesomePrompts
from webscout.AIbase import Provider
from webscout import exceptions


class LearnFast(Provider):
    """
    A class to interact with the LearnFast.ai API.
    """

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        system_prompt: str = "You are a helpful AI assistant.",
    ):
        """
        Initializes the LearnFast.ai API with given parameters.
        """
        self.session = cloudscraper.create_scraper()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = 'https://autosite.erweima.ai/api/v1/chat'
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.system_prompt = system_prompt
        self.headers = {
            "authority": "autosite.erweima.ai",
            "accept": "*/*",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "authorization": "",  # Always empty
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://learnfast.ai",
            "priority": "u=1, i",
            "referer": "https://learnfast.ai/",
            "sec-ch-ua": '"Microsoft Edge";v="129", "Not=A?Brand";v="8", "Chromium";v="129"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "cross-site",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36 Edg/128.0.0.0",
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.session.headers.update(self.headers)
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def generate_unique_id(self) -> str:
        """Generate a 32-character hexadecimal unique ID."""
        return uuid.uuid4().hex

    def generate_session_id(self) -> str:
        """Generate a 32-character hexadecimal session ID."""
        return uuid.uuid4().hex

    def upload_image_to_0x0(self, image_path: str) -> str:
        """
        Uploads an image to 0x0.st and returns the public URL.
        """
        if not os.path.isfile(image_path):
            raise FileNotFoundError(f"The file '{image_path}' does not exist.")

        with open(image_path, "rb") as img_file:
            files = {"file": img_file}
            try:
                response = requests.post("https://0x0.st", files=files)
                response.raise_for_status()
                image_url = response.text.strip()
                if not image_url.startswith("http"):
                    raise ValueError("Received an invalid URL from 0x0.st.")
                return image_url
            except requests.exceptions.RequestException as e:
                raise Exception(f"Failed to upload image to 0x0.st: {e}") from e

    def create_payload(
        self,
        session_id: str,
        conversation_prompt: str,
        image_url: Optional[str] = None
    ) -> dict:
        """
        Creates the JSON payload for the request.
        """
        payload = {
            "prompt": conversation_prompt,
            "sessionId": session_id,
        }
        if image_url:
            payload["attachments"] = [
                {
                    "fileType": "image/jpeg",
                    "file": {},
                    "fileContent": image_url
                }
            ]
        return payload

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
        image_path: Optional[str] = None,
    ) -> dict:
        """Chat with LearnFast

        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            raw (bool, optional): Stream back raw response as received. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
            image_path (Optional[str], optional): Path to the image to be uploaded.
                                                 Defaults to None.

        Returns:
           dict : {}
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )

        # Generate unique ID and session ID
        unique_id = self.generate_unique_id()
        session_id = self.generate_session_id()

        # Update headers with the unique ID
        self.headers["uniqueid"] = unique_id

        # Upload image and get URL if image_path is provided
        image_url = None
        if image_path:
            try:
                image_url = self.upload_image_to_0x0(image_path)
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"Error uploading image: {e}") from e

        # Create the payload
        payload = self.create_payload(session_id, conversation_prompt, image_url)

        # Convert the payload to a JSON string
        data = json.dumps(payload)

        try:
            # Send the POST request with streaming enabled
            response = self.session.post(self.api_endpoint, headers=self.headers, data=data, stream=True, timeout=self.timeout)
            response.raise_for_status()  # Check for HTTP errors

            # Process the streamed response
            full_response = ""
            for line in response.iter_lines(decode_unicode=True):
                if line:
                    if line.strip() == "[DONE]":
                        break
                    try:
                        json_response = json.loads(line)
                        message = json_response.get('data', {}).get('message', '')
                        if message:
                            full_response += message
                            # print(message, end='', flush=True)
                    except json.JSONDecodeError:
                        print(f"\nFailed to parse JSON: {line}")
            self.last_response.update({"text": full_response})
            self.conversation.update_chat_history(prompt, full_response)

            return self.last_response
        except requests.exceptions.RequestException as e:
            raise exceptions.FailedToGenerateResponseError(f"An error occurred: {e}")

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
        image_path: Optional[str] = None,
    ) -> str:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
            image_path (Optional[str], optional): Path to the image to be uploaded.
                                                 Defaults to None.
        Returns:
            str: Response generated
        """
        response = self.ask(prompt, stream, optimizer=optimizer, conversationally=conversationally, image_path=image_path)
        return self.get_message(response)

    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response["text"]

if __name__ == "__main__":
    from rich import print
    ai = LearnFast()
    response = ai.chat(input(">>> "), image_path="photo_2024-07-06_22-19-42.jpg")
    for chunk in response:
        print(chunk, end="", flush=True)


import requests
import re
import json
import yaml
from webscout.AIutel import Optimizers
from webscout.AIutel import Conversation
from webscout.AIutel import AwesomePrompts, sanitize_stream
from webscout.AIbase import  Provider

from webscout import exceptions
from typing import Any, AsyncGenerator, Dict


#------------------------------------------------------phind-------------------------------------------------------------
class PhindSearch:
    # default_model = "Phind Model"
    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 8000,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        model: str = "Phind Model",
        quiet: bool = False,
    ):
        """Instantiates PHIND

        Args:
            is_conversation (bool, optional): Flag for chatting conversationally. Defaults to True
            max_tokens (int, optional): Maximum number of tokens to be generated upon completion. Defaults to 600.
            timeout (int, optional): Http request timeout. Defaults to 30.
            intro (str, optional): Conversation introductory prompt. Defaults to None.
            filepath (str, optional): Path to file containing conversation history. Defaults to None.
            update_file (bool, optional): Add new prompts and responses to the file. Defaults to True.
            proxies (dict, optional): Http request proxies. Defaults to {}.
            history_offset (int, optional): Limit conversation history to this number of last texts. Defaults to 10250.
            act (str|int, optional): Awesome prompt key or index. (Used as intro). Defaults to None.
            model (str, optional): Model name. Defaults to "Phind Model".
            quiet (bool, optional): Ignore web search-results and yield final response only. Defaults to False.
        """
        self.session = requests.Session()
        self.max_tokens_to_sample = max_tokens
        self.is_conversation = is_conversation
        self.chat_endpoint = "https://https.extension.phind.com/agent/"
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.model = model
        self.quiet = quiet

        self.headers = {
            "Content-Type": "application/json",
            "User-Agent": "",
            "Accept": "*/*",
            "Accept-Encoding": "Identity",
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.session.headers.update(self.headers)
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> dict:
        """Chat with AI

        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            raw (bool, optional): Stream back raw response as received. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
           dict : {}
        ```json
        {
            "id": "chatcmpl-r0wujizf2i2xb60mjiwt",
            "object": "chat.completion.chunk",
            "created": 1706775384,
            "model": "trt-llm-phind-model-serving",
            "choices": [
                {
                    "index": 0,
                    "delta": {
                        "content": "Hello! How can I assist you with your programming today?"
                        },
                    "finish_reason": null
                }
            ]
        }
        ```
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )

        self.session.headers.update(self.headers)
        payload = {
            "additional_extension_context": "",
            "allow_magic_buttons": True,
            "is_vscode_extension": True,
            "message_history": [
                {"content": conversation_prompt, "metadata": {}, "role": "user"}
            ],
            "requested_model": self.model,
            "user_input": prompt,
        }

        def for_stream():
            response = self.session.post(
                self.chat_endpoint, json=payload, stream=True, timeout=self.timeout
            )
            if (
                not response.ok
                or not response.headers.get("Content-Type")
                == "text/event-stream; charset=utf-8"
            ):
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                )
            streaming_text = ""
            for value in response.iter_lines(
                decode_unicode=True,
                chunk_size=self.stream_chunk_size,
            ):
                try:
                    modified_value = re.sub("data:", "", value)
                    json_modified_value = json.loads(modified_value)
                    retrieved_text = self.get_message(json_modified_value)
                    if not retrieved_text:
                        continue
                    streaming_text += retrieved_text
                    json_modified_value["choices"][0]["delta"][
                        "content"
                    ] = streaming_text
                    self.last_response.update(json_modified_value)
                    yield value if raw else json_modified_value
                except json.decoder.JSONDecodeError:
                    pass
            self.conversation.update_chat_history(
                prompt, self.get_message(self.last_response)
            )

        def for_non_stream():
            for _ in for_stream():
                pass
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
            str: Response generated
        """

        def for_stream():
            for response in self.ask(
                prompt, True, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            return self.get_message(
                self.ask(
                    prompt,
                    False,
                    optimizer=optimizer,
                    conversationally=conversationally,
                )
            )

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        if response.get("type", "") == "metadata":
            return

        delta: dict = response["choices"][0]["delta"]

        if not delta:
            return ""

        elif delta.get("function_call"):
            if self.quiet:
                return ""

            function_call: dict = delta["function_call"]
            if function_call.get("name"):
                return function_call["name"]
            elif function_call.get("arguments"):
                return function_call.get("arguments")

        elif delta.get("metadata"):
            if self.quiet:
                return ""
            return yaml.dump(delta["metadata"])

        else:
            return (
                response["choices"][0]["delta"].get("content")
                if response["choices"][0].get("finish_reason") is None
                else ""
            )

class Phindv2(Provider):
    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 8000,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        model: str = "Phind Instant",
        quiet: bool = False,
        system_prompt: str = "Be Helpful and Friendly",
    ):
        """Instantiates Phindv2

        Args:
            is_conversation (bool, optional): Flag for chatting conversationally. Defaults to True.
            max_tokens (int, optional): Maximum number of tokens to be generated upon completion. Defaults to 600.
            timeout (int, optional): Http request timeout. Defaults to 30.
            intro (str, optional): Conversation introductory prompt. Defaults to None.
            filepath (str, optional): Path to file containing conversation history. Defaults to None.
            update_file (bool, optional): Add new prompts and responses to the file. Defaults to True.
            proxies (dict, optional): Http request proxies. Defaults to {}.
            history_offset (int, optional): Limit conversation history to this number of last texts. Defaults to 10250.
            act (str|int, optional): Awesome prompt key or index. (Used as intro). Defaults to None.
            model (str, optional): Model name. Defaults to "Phind Model".
            quiet (bool, optional): Ignore web search-results and yield final response only. Defaults to False.
            system_prompt (str, optional): System prompt for Phindv2. Defaults to "Be Helpful and Friendly".
        """
        self.session = requests.Session()
        self.max_tokens_to_sample = max_tokens
        self.is_conversation = is_conversation
        self.chat_endpoint = "https://https.extension.phind.com/agent/"
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.model = model
        self.quiet = quiet
        self.system_prompt = system_prompt

        self.headers = {
            "Content-Type": "application/json",
            "User-Agent": "",
            "Accept": "*/*",
            "Accept-Encoding": "Identity",
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.session.headers.update(self.headers)
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> dict:
        """Chat with AI

        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            raw (bool, optional): Stream back raw response as received. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
           dict : {}
        ```json
        {
            "id": "chatcmpl-r0wujizf2i2xb60mjiwt",
            "object": "chat.completion.chunk",
            "created": 1706775384,
            "model": "trt-llm-phind-model-serving",
            "choices": [
                {
                    "index": 0,
                    "delta": {
                        "content": "Hello! How can I assist you with your programming today?"
                        },
                    "finish_reason": null
                }
            ]
        }
        ```
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )

        self.session.headers.update(self.headers)
        payload = {
            "additional_extension_context": "",
            "allow_magic_buttons": True,
            "is_vscode_extension": True,
            "message_history": [
                {"content": self.system_prompt, "metadata": {}, "role": "system"},
                {"content": conversation_prompt, "metadata": {}, "role": "user"}
            ],
            "requested_model": self.model,
            "user_input": prompt,
        }

        def for_stream():
            response = self.session.post(
                self.chat_endpoint, json=payload, stream=True, timeout=self.timeout
            )
            if (
                not response.ok
                or not response.headers.get("Content-Type")
                == "text/event-stream; charset=utf-8"
            ):
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                )
            streaming_text = ""
            for value in response.iter_lines(
                decode_unicode=True,
                chunk_size=self.stream_chunk_size,
            ):
                try:
                    modified_value = re.sub("data:", "", value)
                    json_modified_value = json.loads(modified_value)
                    retrieved_text = self.get_message(json_modified_value)
                    if not retrieved_text:
                        continue
                    streaming_text += retrieved_text
                    json_modified_value["choices"][0]["delta"][
                        "content"
                    ] = streaming_text
                    self.last_response.update(json_modified_value)
                    yield value if raw else json_modified_value
                except json.decoder.JSONDecodeError:
                    pass
            self.conversation.update_chat_history(
                prompt, self.get_message(self.last_response)
            )

        def for_non_stream():
            for _ in for_stream():
                pass
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
            str: Response generated
        """

        def for_stream():
            for response in self.ask(
                prompt, True, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            return self.get_message(
                self.ask(
                    prompt,
                    False,
                    optimizer=optimizer,
                    conversationally=conversationally,
                )
            )

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        if response.get("type", "") == "metadata":
            return

        delta: dict = response["choices"][0]["delta"]

        if not delta:
            return ""

        elif delta.get("function_call"):
            if self.quiet:
                return ""

            function_call: dict = delta["function_call"]
            if function_call.get("name"):
                return function_call["name"]
            elif function_call.get("arguments"):
                return function_call.get("arguments")

        elif delta.get("metadata"):
            if self.quiet:
                return ""
            return yaml.dump(delta["metadata"])

        else:
            return (
                response["choices"][0]["delta"].get("content")
                if response["choices"][0].get("finish_reason") is None
                else ""
            )




import requests
import json
from typing import Any, Dict, Optional
from webscout.AIutel import Optimizers
from webscout.AIutel import Conversation
from webscout.AIutel import AwesomePrompts, sanitize_stream
from webscout.AIbase import Provider
from webscout import exceptions

class KOALA(Provider):
    """
    A class to interact with the Koala.sh API.
    """

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        model: str = "gpt-4o-mini",
        web_search: bool = True,

    ) -> None:
        """
        Initializes the KOALASH API with given parameters.

        Args:
            is_conversation (bool, optional): Flag for chatting conversationally. Defaults to True.
            max_tokens (int, optional): Maximum number of tokens to be generated upon completion. 
                                        Defaults to 600.
            timeout (int, optional): Http request timeout. Defaults to 30.
            intro (str, optional): Conversation introductory prompt. Defaults to None.
            filepath (str, optional): Path to file containing conversation history. Defaults to None.
            update_file (bool, optional): Add new prompts and responses to the file. Defaults to True.
            proxies (dict, optional): Http request proxies. Defaults to {}.
            history_offset (int, optional): Limit conversation history to this number of last texts. 
                                            Defaults to 10250.
            act (str|int, optional): Awesome prompt key or index. (Used as intro). Defaults to None.
            model (str, optional): AI model to use. Defaults to "gpt-4o-mini".
        """
        self.session = requests.Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = "https://koala.sh/api/gpt/"
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.model = model
        self.headers = {
            "accept": "text/event-stream",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-length": "73",
            "content-type": "application/json",
            "dnt": "1",
            "flag-real-time-data": "true" if web_search else "false", 
            "origin": "https://koala.sh",
            "priority": "u=1, i",
            "referer": "https://koala.sh/chat",
            "sec-ch-ua": '"Not)A;Brand";v="99", "Microsoft Edge";v="127", "Chromium";v="127"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36 Edg/127.0.0.0",
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.session.headers.update(self.headers)
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> Dict[str, Any]:
        """
        Sends a prompt to the Koala.sh API and returns the response.

        Args:
            prompt: The text prompt to generate text from.
            stream (bool, optional): Whether to stream the response. Defaults to False.
            raw (bool, optional): Whether to return the raw response. Defaults to False.
            optimizer (str, optional): The name of the optimizer to use. Defaults to None.
            conversationally (bool, optional): Whether to chat conversationally. Defaults to False.

        Returns:
            The response from the API.
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )
        
        payload = {
            "input": conversation_prompt,
            "model": self.model 
        }

        def for_stream():
            response = self.session.post(
                self.api_endpoint, json=payload, headers=self.headers, stream=True, timeout=self.timeout
            )

            if not response.ok:
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to generate response - ({response.status_code}, {response.reason})"
                )
            
            streaming_response = ""
            for line in response.iter_lines(decode_unicode=True):
                if line:
                    if line.startswith("data:"):
                        data = line[len("data:"):].strip()
                        if data:
                            try:
                                event = json.loads(data)
                                streaming_response += event.get("choices", [{}])[0].get("delta", {}).get("content", "")
                                yield event if raw else dict(text=streaming_response)
                            except json.decoder.JSONDecodeError:
                                continue
            self.last_response.update(dict(text=streaming_response))
            self.conversation.update_chat_history(
                prompt, self.get_message(self.last_response)
            )
        def for_non_stream():
            response = self.session.post(
                self.api_endpoint, json=payload, headers=self.headers, timeout=self.timeout
            )

            if not response.ok:
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to generate response - ({response.status_code}, {response.reason})"
                )
            
            response_content = response.content.decode('utf-8')
            data_parts = response_content.strip().split('\n\n')
            formatted_response = ''.join([part.replace('data: ', '') for part in data_parts if part.startswith('data: ')])

            # Remove extra quotes from the formatted response
            formatted_response = formatted_response.replace('""', '')

            # Split the response into lines and format with new lines before headers
            lines = formatted_response.split('\n')
            formatted_lines = []
            for line in lines:
                if line.startswith('###'):
                    formatted_lines.append('\n' + line)
                else:
                    formatted_lines.append(line)

            # Join the formatted lines back into a single string
            final_response = '\n'.join(formatted_lines)

            # self.last_response.update(dict(text=streaming_response))
            self.conversation.update_chat_history(
                prompt, final_response
            )
            return dict(text=final_response)

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
            str: Response generated
        """

        def for_stream():
            for response in self.ask(
                prompt, True, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            return self.get_message(
                self.ask(
                    prompt,
                    False,
                    optimizer=optimizer,
                    conversationally=conversationally,
                )
            )

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response["text"].replace('\\n', '\n').replace('\\n\\n', '\n\n')
if __name__ == '__main__':
    from rich import print
    ai = KOALA()
    response = ai.chat("tell me about india")
    for chunk in response:
        print(chunk, end="", flush=True)


import os
import sys
import json
import base64

from httpx import stream

from .model import Model

from .utils import (
    InferenceLock,
    download_model,
    print_verbose,
    SPECIAL_STYLE,
    assert_type,
    ERROR_STYLE,
    USER_STYLE,
    BOT_STYLE,
    RESET_ALL
)

from .thread import Thread
from cryptography import x509
from cryptography.x509.oid import NameOID
from datetime import datetime, timedelta, UTC
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from flask import Flask, logging, render_template, request, Response
from cryptography.hazmat.primitives.serialization import Encoding

# Color codes for console output
YELLOW = SPECIAL_STYLE
GREEN = USER_STYLE
RED = ERROR_STYLE
RESET = RESET_ALL
BLUE = BOT_STYLE

# Warning message for WebUI security
WARNING = f"""{RED}
################################################################################
{RESET}

                          PLEASE KEEP IN MIND

        The webscout.Local WebUI is not guaranteed to be secure.

             It is not intended to be exposed to the internet.

     If you expose the WebUI to the internet, you do so at your own risk.

                          YOU HAVE BEEN WARNED!

{RED}
################################################################################
{RESET}"""

# SSL certificate generation warning
SSL_CERT_FIRST_TIME_WARNING = (
    f"{YELLOW}You have just generated a new self-signed SSL certificate and "
    f"key. Your browser will probably warn you about an untrusted "
    f"certificate. This is expected, and you may safely proceed to the WebUI. "
    f"Subsequent WebUI sessions will reuse this SSL certificate.{RESET}"
)

# Constants
ASSETS_FOLDER = os.path.join(os.path.dirname(__file__), "assets")
MAX_LENGTH_INPUT = 100_000  # Maximum input length (characters)

def generate_self_signed_ssl_cert() -> None:
    """Generates a self-signed SSL certificate and key."""
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    public_key = private_key.public_key()

    # Certificate details
    name = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "XY"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "DUMMY_STATE"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "DUMMY_LOCALITY"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "EZLLAMA LLC"),
        x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
    ])

    # Certificate builder
    builder = x509.CertificateBuilder(
        subject_name=name,
        issuer_name=name,
        public_key=public_key,
        serial_number=x509.random_serial_number(),
        not_valid_before=datetime.now(tz=UTC),
        not_valid_after=datetime.now(tz=UTC) + timedelta(days=36500),
    )
    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName("localhost")]),
        critical=False,
    )

    # Sign and save the certificate and key
    certificate = builder.sign(private_key=private_key, algorithm=hashes.SHA256())
    with open(f"{ASSETS_FOLDER}/key.pem", "wb") as f:
        f.write(private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        ))
    with open(f"{ASSETS_FOLDER}/cert.pem", "wb") as f:
        f.write(certificate.public_bytes(Encoding.PEM))

def check_for_ssl_cert() -> bool:
    """Checks if SSL certificate and key exist."""
    return all(os.path.exists(f"{ASSETS_FOLDER}/{file}.pem") for file in ["cert", "key"])

def newline() -> None:
    """Prints a newline to stderr."""
    print("", end="\n", file=sys.stderr, flush=True)

def encode(text: str) -> str:
    """Encodes a string to base64."""
    return base64.b64encode(text.encode("utf-8")).decode("utf-8")

def decode(text: str) -> str:
    """Decodes a base64 encoded string."""
    return base64.b64decode(text).decode("utf-8")

def assert_max_length(text: str) -> None:
    """Asserts that the length of the text is within the allowed limit."""
    if len(text) > MAX_LENGTH_INPUT:
        raise AssertionError(
            f"Input text exceeds maximum length of {MAX_LENGTH_INPUT} characters."
        )

def _print_inference_string(text: str) -> None:
    """Prints the inference string to stderr."""
    print(
        f"{'#' * 80}\n"
        f"{YELLOW}'''{RESET}{text}{YELLOW}'''{RESET}\n"
        f"{'#' * 80}",
        file=sys.stderr,
        flush=True
    )


class WebUI:
    """
    Represents the webscout.Local WebUI server.
    """

    def __init__(self, thread: Thread):
        """
        Initializes the WebUI instance.

        Args:
            thread (Thread): The Thread instance to use for the WebUI.
        """
        assert_type(thread, Thread, "thread", "WebUI")
        self.thread = thread
        self.lock = InferenceLock()
        self._cancel_flag = False

        self.app = Flask(
            __name__,
            static_folder=ASSETS_FOLDER,
            template_folder=ASSETS_FOLDER,
            static_url_path="",
        )

        self._log_host = None
        self._log_port = None

    def log(self, text: str) -> None:
        """Logs a message to the console."""
        if self._log_host is None or self._log_port is None:
            print_verbose(text)
        else:
            print(
                f"webscout.Local: WebUI @ "
                f"{YELLOW}{self._log_host}{RESET}:"
                f"{YELLOW}{self._log_port}{RESET}: {text}",
                file=sys.stderr,
                flush=True
            )

    def _get_context_string(self) -> str:
        """Returns a string representing the current context usage."""
        thread_len_tokens = self.thread.len_messages()
        max_ctx_len = self.thread.model.context_length
        return f"{thread_len_tokens} / {max_ctx_len} tokens used"

    def start(self, host: str, port: int = 8080, ssl: bool = False) -> None:
        """
        Starts the WebUI server.

        Args:
            host (str): The hostname or IP address to bind the server to.
            port (int, optional): The port to listen on. Defaults to 8080.
            ssl (bool, optional): Whether to use SSL. Defaults to False.
        """
        print(WARNING, file=sys.stderr, flush=True)

        assert_type(host, str, "host", "WebUI.start")
        assert_type(port, int, "port", "WebUI.start")

        self._log_host = host
        self._log_port = port

        self.log(f"Starting WebUI instance:")
        self.log(f"   thread.uuid == {self.thread.uuid}")
        self.log(f"   host        == {host}")
        self.log(f"   port        == {port}")
        self.log(f"   ssl (HTTPS) == {ssl}")

        if ssl:
            if not check_for_ssl_cert():
                self.log("Generating self-signed SSL certificate...")
                generate_self_signed_ssl_cert()
                print_verbose(SSL_CERT_FIRST_TIME_WARNING)
            else:
                self.log("Reusing previously generated SSL certificate.")

        # Flask route definitions
        @self.app.route("/")
        def home():
            return render_template("index.html")

        @self.app.route("/convo", methods=["GET"])
        def convo():
            msgs_dict = {i: {encode(msg["role"]): encode(msg["content"])}
                         for i, msg in enumerate(self.thread.messages)}
            json_convo = json.dumps(msgs_dict)
            return json_convo, 200, {"ContentType": "application/json"}

        @self.app.route("/cancel", methods=["POST"])
        def cancel():
            newline()
            self.log("Hit cancel endpoint - flag is set.")
            self._cancel_flag = True
            return "", 200

        @self.app.route("/submit", methods=["POST"])
        def submit():
            self.log("Hit submit endpoint.")
            prompt = decode(request.data)
            assert_max_length(prompt)

            if not prompt:
                self.log("Empty prompt submitted. Ignoring.")
                return "", 200

            # Pass the stream variable to the generate function
            def generate(stream=stream): 
                with self.lock:
                    self.thread.add_message("user", prompt)
                    print(f"{GREEN}{prompt}{RESET}", file=sys.stderr)
                    inf_str = self.thread.inference_str_from_messages()
                    _print_inference_string(inf_str)

                    if stream:
                        token_generator = self.thread.model.stream(
                            inf_str,
                            stops=self.thread.format["stops"],
                            sampler=self.thread.sampler
                        )
                        response = ""
                        for token in token_generator:
                            if self._cancel_flag:
                                print(file=sys.stderr)
                                self.log("Canceling generation from /submit.")
                                self._cancel_flag = False
                                return "", 418  # I'm a teapot
                            
                            tok_text = token['choices'][0]['text']
                            response += tok_text
                            print(f'{BLUE}{tok_text}{RESET}', end='', flush=True, file=sys.stderr)
                            yield encode(tok_text) + "\n"
                    else:
                        response = self.thread.model.generate(
                            inf_str,
                            stops=self.thread.format["stops"],
                            sampler=self.thread.sampler
                        )
                        # Simulate streaming by yielding chunks of the content
                        chunk_size = self.stream_chunk_size
                        for i in range(0, len(response), chunk_size):
                            chunk = response[i:i + chunk_size]
                            yield encode(chunk) + "\n"

                    self._cancel_flag = False
                    newline()
                    self.thread.add_message("bot", response)

            return Response(generate(), mimetype="text/plain")

        @self.app.route("/reset", methods=["POST"])
        def reset():
            self.thread.reset()
            self.log("Thread reset.")
            return "", 200

        @self.app.route("/get_context_string", methods=["GET"])
        def get_context_string():
            return json.dumps({"text": encode(self._get_context_string())}), 200, {"ContentType": "application/json"}

        @self.app.route("/remove", methods=["POST"])
        def remove():
            if len(self.thread.messages) > 1:  # Prevent deleting the system message
                self.thread.messages.pop(-1)
                self.log("Removed last message.")
                return "", 200
            else:
                self.log("No previous message to remove.")
                return "", 418  # I'm a teapot

        @self.app.route("/trigger", methods=["POST"])
        def trigger():
            self.log("Hit trigger endpoint.")
            prompt = decode(request.data)
            assert_max_length(prompt)

            if prompt:
                self.log(f"Trigger with prompt: {prompt!r}")
            else:
                self.log("Trigger without prompt.")

            # Pass stream to the generate function
            def generate(stream=stream):
                with self.lock:
                    inf_str = self.thread.inference_str_from_messages() + prompt
                    _print_inference_string(inf_str)

                    if stream:
                        token_generator = self.thread.model.stream(
                            inf_str,
                            stops=self.thread.format["stops"],
                            sampler=self.thread.sampler
                        )
                        response = ""
                        for token in token_generator:
                            if self._cancel_flag:
                                print(file=sys.stderr)
                                self.log("Canceling generation from /trigger.")
                                self._cancel_flag = False
                                return "", 418  # I'm a teapot

                            tok_text = token["choices"][0]["text"]
                            response += tok_text
                            print(f"{BLUE}{tok_text}{RESET}", end='', flush=True)
                            yield encode(tok_text) + "\n"
                    else:
                        response = self.thread.model.generate(
                            inf_str,
                            stops=self.thread.format["stops"],
                            sampler=self.thread.sampler
                        )
                        # Simulate streaming by yielding chunks of the content
                        chunk_size = self.stream_chunk_size
                        for i in range(0, len(response), chunk_size):
                            chunk = response[i:i + chunk_size]
                            yield encode(chunk) + "\n"

                    self._cancel_flag = False
                    print("", file=sys.stderr)
                    self.thread.add_message("bot", prompt + response)

            return Response(generate(), mimetype="text/plain")

        @self.app.route("/summarize", methods=["GET"])
        def summarize():
            with self.lock:
                summary = self.thread.summarize()
            self.log(f"Generated summary: {BLUE}{summary!r}{RESET}")
            return encode(summary), 200, {"ContentType": "text/plain"}

        if not self.thread.model.is_loaded():
            self.log("Loading model...")
            self.thread.model.load()
        else:
            self.log("Model is already loaded.")

        self.log("Warming up thread...")
        self.thread.warmup()

        self.log("Running Flask app...")
        try:
            self.app.run(
                host=host,
                port=port,
                ssl_context=(
                    (f"{ASSETS_FOLDER}/cert.pem", f"{ASSETS_FOLDER}/key.pem")
                    if ssl
                    else None
                ),
            )
        except Exception as exc:
            newline()
            self.log(f"{RED}Exception in Flask: {exc}{RESET}")
            raise exc
        else:
            newline()
            self._log_host = None
            self._log_port = None




import requests
import base64
from typing import List, Dict, Union
import json
import requests
import base64
from typing import List, Dict, Union

class LLM:
    def __init__(self, model: str, system_message: str = "You are a Helpful AI."):
        self.model = model
        self.conversation_history = [{"role": "system", "content": system_message}]

    def chat(self, messages: List[Dict[str, str]]) -> Union[str, None]:
        url = "https://api.deepinfra.com/v1/openai/chat/completions"
        headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36',
            'Accept-Language': 'en,fr-FR;q=0.9,fr;q=0.8,es-ES;q=0.7,es;q=0.6,en-US;q=0.5,am;q=0.4,de;q=0.3',
            'Cache-Control': 'no-cache',
            'Connection': 'keep-alive',
            'Content-Type': 'application/json',
            'Origin': 'https://deepinfra.com',
            'Pragma': 'no-cache',
            'Referer': 'https://deepinfra.com/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-site',
            'X-Deepinfra-Source': 'web-embed',
            'accept': 'text/event-stream',
            'sec-ch-ua': '"Google Chrome";v="119", "Chromium";v="119", "Not?A_Brand";v="24"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"macOS"'
        }
        data = json.dumps(
            {
                'model': self.model,
                'messages': messages,
                'temperature': 0.7,
                'max_tokens': 8028,
                'stop': [],
                'stream': False #dont change it
            }, separators=(',', ':')
        )
        try:
            result = requests.post(url=url, data=data, headers=headers)
            return result.json()['choices'][0]['message']['content']
        except:
            return None
# def main():
#     llm = LLM(model="meta-llama/Meta-Llama-3-70B-Instruct")
#     messages = [
#         {"role": "user", "content": "Hello, how are you?"}
#     ]
#     response = llm.chat(messages)
#     print(response)

# if __name__ == "__main__":
#     main()


class VLM:
    def __init__(self, model: str, system_message: str = "You are a Helpful AI."):
        self.model = model
        self.conversation_history = [{"role": "system", "content": system_message}]

    def chat(self, messages: List[Dict[str, Union[str, List[Dict[str, Union[str, Dict[str, str]]]]]]]) -> Union[str, None]:
        api_url = "https://api.deepinfra.com/v1/openai/chat/completions"
        headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36',
            'Accept-Language': 'en,fr-FR;q=0.9,fr;q=0.8,es-ES;q=0.7,es;q=0.6,en-US;q=0.5,am;q=0.4,de;q=0.3',
            'Cache-Control': 'no-cache',
            'Connection': 'keep-alive',
            'Content-Type': 'application/json',
            'Origin': 'https://deepinfra.com',
            'Pragma': 'no-cache',
            'Referer': 'https://deepinfra.com/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-site',
            'X-Deepinfra-Source': 'web-embed',
            'accept': 'text/event-stream',
            'sec-ch-ua': '"Google Chrome";v="119", "Chromium";v="119", "Not?A_Brand";v="24"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"macOS"'
        }
        payload = {
            "model": self.model,
            "messages": messages,
            "stream": False
        }
        try:
            response = requests.post(api_url, headers=headers, json=payload)
            return response.json()['choices'][0]['message']['content']
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

def encode_image_to_base64(image_path: str) -> str:
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")



import json
import requests
from webscout.AIutel import Optimizers, Conversation, AwesomePrompts
from webscout.AIbase import Provider
from webscout import exceptions
from typing import Dict, Any

class Cerebras(Provider):
    """
    A class to interact with the Cerebras AI API.
    """

    AVAILABLE_MODELS = ["llama3.1-8b", "llama3.1-70b"]

    def __init__(
        self,
        api_key: str,
        is_conversation: bool = True,
        max_tokens: int = 4096,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        model: str = "llama3.1-8b",
        system_prompt: str = "Please try to provide useful, helpful and actionable answers.",
    ):
        """
        Initializes the Cerebras AI API with given parameters.
        """
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Available models are: {', '.join(self.AVAILABLE_MODELS)}")

        self.session = requests.Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = "https://api.cerebras.ai/v1/chat/completions"
        self.timeout = timeout
        self.last_response = {}
        self.model = model
        self.system_prompt = system_prompt
        self.headers = {
            "accept": "application/json",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "authorization": f"Bearer {api_key}",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://inference.cerebras.ai",
            "referer": "https://inference.cerebras.ai/",
            "sec-ch-ua": '"Chromium";v="128", "Not;A=Brand";v="24", "Microsoft Edge";v="128"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-site",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36 Edg/128.0.0.0",
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.session.headers.update(self.headers)
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> Dict[str, Any]:
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )

        payload = {
            "messages": [
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": conversation_prompt},
            ],
            "model": self.model,
            "stream": True,
            "temperature": 0.2,
            "top_p": 1,
            "max_tokens": self.max_tokens_to_sample
        }

        def for_stream():
            response = self.session.post(
                self.api_endpoint, json=payload, stream=True, timeout=self.timeout
            )

            if not response.ok:
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to generate response - ({response.status_code}, {response.reason})"
                )

            full_response = ""
            for line in response.iter_lines():
                if line:
                    line_data = line.decode('utf-8').strip()
                    if line_data.startswith("data: "):
                        json_str = line_data[6:]
                        if json_str != "[DONE]":
                            chunk = json.loads(json_str)
                            if 'choices' in chunk and 'delta' in chunk['choices'][0]:
                                content = chunk['choices'][0]['delta'].get('content', '')
                                full_response += content
                                yield content if raw else dict(text=content)
                        else:
                            break

            self.last_response.update(dict(text=full_response))
            self.conversation.update_chat_history(
                prompt, self.get_message(self.last_response)
            )

        def for_non_stream():
            full_response = ""
            for chunk in for_stream():
                if isinstance(chunk, dict):
                    full_response += chunk['text']
                else:
                    full_response += chunk
            return dict(text=full_response)

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str:
        def for_stream():
            for response in self.ask(
                prompt, True, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            return self.get_message(
                self.ask(
                    prompt,
                    False,
                    optimizer=optimizer,
                    conversationally=conversationally,
                )
            )

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response["text"]

if __name__ == '__main__':
    from rich import print
    
    # You can replace this with your actual API key
    api_key = "YOUR_API_KEY_HERE"
    
    ai = Cerebras(api_key=api_key)
    response = ai.chat(input(">>> "), stream=True)
    for chunk in response:
        print(chunk, end="", flush=True)



import aiohttp
from dataclasses import dataclass
from bs4 import BeautifulSoup
import tls_client
import random
from typing import ClassVar, NoReturn, List, Dict, Any
import requests

@dataclass
class DomainModel:
    name: str
    type: str
    forward_available: str
    forward_max_seconds: str


@dataclass
class CreateEmailResponseModel:
    email: str
    token: str


@dataclass
class MessageResponseModel:
    attachments: list | None
    body_html: str | None
    body_text: str | None
    cc: str | None
    created_at: str
    email_from: str | None
    id: str
    subject: str | None
    email_to: str | None


class Client:
    def __init__(self):
        self._session = aiohttp.ClientSession(
            base_url="https://api.internal.temp-mail.io",
            headers={
                'Host': 'api.internal.temp-mail.io',
                'User-Agent': 'okhttp/4.5.0',
                'Connection': 'close'
            }
        )

    async def close(self) -> None:
        if not self._session.closed:
            await self._session.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self) -> None:
        await self.close()
        return None

    async def get_domains(self) -> list[DomainModel]:
        async with self._session.get("/api/v3/domains") as response:
            response_json = await response.json()
            return [DomainModel(domain['name'], domain['type'], domain['forward_available'], domain['forward_max_seconds']) for domain in response_json['domains']]

    async def create_email(self, alias: str | None = None, domain: str | None = None) -> CreateEmailResponseModel:
        async with self._session.post("/api/v3/email/new", data={'name': alias, 'domain': domain}) as response:
            response_json = await response.json()
            return CreateEmailResponseModel(response_json['email'], response_json['token'])

    async def delete_email(self, email: str, token: str) -> bool:
        async with self._session.delete(f"/api/v3/email/{email}", data={'token': token}) as response:
            if response.status == 200:
                return True
            else:
                return False

    async def get_messages(self, email: str) -> list[MessageResponseModel] | None:
        async with self._session.get(f"/api/v3/email/{email}/messages") as response:
            response_json = await response.json()
            if len(response_json) == 0:
                return None
            return [MessageResponseModel(message['attachments'], message['body_html'], message['body_text'], message['cc'], message['created_at'], message['from'], message['id'], message['subject'], message['to']) for message in response_json]


class TemporaryPhoneNumber:
    def __init__(self):
        self.maxpages = {"UK": 59, "US": 3, "France": 73, "Netherlands": 60, "Finland": 47}
        self.minpages = {"UK": 20, "US": 1, "France": 20, "Netherlands": 20, "Finland": 20}
        self.plist = {"UK": "+44", "US": "+1", "France": "+33", "Netherlands": "+31", "Finland": "+358"}
        self.countries = {"44": "UK", "1": "US", "33": "France", "31": "Netherlands", "358": "Finland"}

    def get_number(self, country="UK"):
        if country == "Random":
            country = random.choice(list(self.countries.values()))
        if country not in self.countries.values():
            raise ValueError("Unsupported Country")

        session = tls_client.Session(client_identifier="chrome112", random_tls_extension_order=True)
        maxpage = self.maxpages[country]
        minpage = self.minpages[country]
        page = random.randint(minpage, maxpage)

        if page == 1:
            res = session.get(f"https://temporary-phone-number.com/{country}-Phone-Number")
        else:
            res = session.get(f"https://temporary-phone-number.com/{country}-Phone-Number/page{page}")

        soup = BeautifulSoup(res.content, "lxml")
        numbers = []
        p = self.plist[country]
        for a in soup.find_all("a"):
            a = a.get("title", "none")
            if f"{country} Phone Number {p}" in a:
                a = a.replace(f"{country} Phone Number ", "").replace(" ", "")
                numbers.append(a)
        return random.choice(numbers)

    def get_messages(self, number: str):
        number = number.replace("+", "")
        try:
            i = int(number)
        except:
            raise ValueError("Wrong Number")

        country = None
        for key, value in self.countries.items():
            if number.startswith(key):
                country = value

        if country == None:
            raise ValueError("Unsupported Country")

        session = tls_client.Session(client_identifier="chrome112", random_tls_extension_order=True)
        res = session.get(f"https://temporary-phone-number.com/{country}-Phone-Number/{number}")

        if res.status_code == 404:
            raise ValueError("Number doesn't exist")

        soup = BeautifulSoup(res.content, "lxml")
        messages = []
        message = {"content": None, "frm": "", "time": ""}

        for div in soup.find_all("div"):
            divclass = div.get("class", "None")[0]
            if divclass == "direct-chat-info":
                message["frm"] = div.text.split("\n")[1].replace("From ", "")
                message["time"] = div.text.split("\n")[2]
            if divclass == "direct-chat-text":
                message["content"] = div.text
                messages.append(sms_message(content=message["content"], frm=message["frm"], time=message["time"]))
                message = {"content": None, "frm": "", "time": ""}

        return messages
    
class VNEngine:
    def __init__(self) -> NoReturn:
        self.lang: str = "?lang=en"
        self.base: str = "https://onlinesim.io/"
        self.endpoint: str = "api/v1/free_numbers_content/"
        self.country_url: str = f"{self.base}{self.endpoint}countries"

    def get_online_countries(self) -> List[Dict[str, str]]:
        response: Any = requests.get(url=self.country_url).json()
        if response["response"] == "1":
            all_countries: List[Dict[str, str]] = response["counties"]
            online_countries: List[Dict[str, str]] = list(
                filter(lambda x: x["online"] == True, all_countries)
            )
            return online_countries
        return []

    def get_country_numbers(self, country: str) -> List[Dict[str, str]]:
        numbers_url: str = f"{self.country_url}/{country}{self.lang}"
        response: Any = requests.get(url=numbers_url).json()
        if response["response"] == "1":
            numbers: List[Dict[str, str]] = list(
                map(lambda x: {"data_humans": x["data_humans"], "full_number": x["full_number"]}, response["numbers"])
            )
            return numbers
        return []

    def get_number_inbox(self, country: str, number: str) -> Dict[str, str]:
        number_detail_url: str = f"{self.country_url}/{country}/{number}{self.lang}"
        response: Any = requests.get(url=number_detail_url).json()
        if response["response"] != "1" or not response["online"]:
            print(f"Error: Unable to retrieve inbox messages for {country} - {number}")
            return {}

        messages: List[Dict[str, str]] = []
        for msg_data in response["messages"]["data"]:
            try:
                msg = {"data_humans": msg_data["data_humans"], "text": msg_data["text"]}
                messages.append(msg)
            except KeyError as e:
                print(f"Warning: Missing key '{str(e)}' in message data")

        return {"messages": messages}

class sms_message:
    def __init__(self, content, frm, time):
        self.content = content
        self.frm = frm
        self.time = time


import requests
import json
import os
from typing import Any, Dict, Optional, Generator, Union

from webscout.AIutel import Optimizers
from webscout.AIutel import Conversation
from webscout.AIutel import AwesomePrompts, sanitize_stream
from webscout.AIbase import Provider, AsyncProvider
from webscout import exceptions

class Bagoodex(Provider):
    """
    A class to interact with the Bagoodex API.
    """

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 2049,  # Set a reasonable default
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
    ):
        """Initializes the Bagoodex API client."""
        self.url = "https://bagoodex.io/front-api/chat"
        self.headers = {"Content-Type": "application/json"}
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        self.session.proxies.update(proxies)  # Use provided proxies
        self.timeout = timeout
        self.last_response = {}

        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> Dict[str, Any] | Generator:
        """Sends a chat completion request to the Bagoodex API."""

        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")


        payload = {
            "prompt": "You are AI",  # This seems to be required by the API
            "messages": [{"content": "Hi, this is chatgpt, let's talk", "role": "assistant"}],
            "input": conversation_prompt,
        }

        def for_stream():
            try:
                response = self.session.post(self.url, json=payload, headers=self.headers, timeout=self.timeout)
                response.raise_for_status()
                text = response.text
                self.last_response.update({"text": text})

                if stream:
                    for char in text:
                        yield {"text": char}  # Yielding one character at a time for streaming
                else:
                    yield {"text": text}

            except (requests.exceptions.RequestException, json.JSONDecodeError) as e:  # Catch JSON errors too
                raise exceptions.FailedToGenerateResponseError(f"Error during request: {e}")
            self.conversation.update_chat_history(prompt, self.get_message(self.last_response))

        def for_non_stream():
            for _ in for_stream(): pass
            return self.last_response


        return for_stream() if stream else for_non_stream()




    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> Union[str, Generator]:


        def for_stream():
            for response in self.ask(
                prompt, stream=True, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            return self.get_message(
                self.ask(
                    prompt, stream=False, optimizer=optimizer, conversationally=conversationally
                )
            )

        return for_stream() if stream else for_non_stream()


    def get_message(self, response: dict) -> str:
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response.get("text", "")


if __name__ == "__main__":
    from rich import print
    ai = Bagoodex()
    response = ai.chat("write a poem about AI", stream=True)
    for chunk in response:
        print(chunk, end="", flush=True)


import json
import time
from typing import Iterable, Dict, Any, Generator

from os import listdir
from uuid import uuid4
from time import sleep, time
from threading import Thread
from json import loads, dumps
from random import getrandbits
from websocket import WebSocketApp
from requests import Session, get, post


from webscout.AIutel import Optimizers
from webscout.AIutel import Conversation
from webscout.AIutel import AwesomePrompts, sanitize_stream
from webscout.AIbase import Provider, AsyncProvider
from webscout import exceptions


class Perplexity(Provider):
    def __init__(
        self,
        email: str = None,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        quiet: bool = False,
    ) -> None:
        """Instantiates PERPLEXITY

        Args:
            email (str, optional): Your perplexity.ai email. Defaults to None.
            is_conversation (bool, optional): Flag for chatting conversationally. Defaults to True.
            max_tokens (int, optional): Maximum number of tokens to be generated upon completion. Defaults to 600.
            timeout (int, optional): Http request timeout. Defaults to 30.
            intro (str, optional): Conversation introductory prompt. Defaults to None.
            filepath (str, optional): Path to file containing conversation history. Defaults to None.
            update_file (bool, optional): Add new prompts and responses to the file. Defaults to True.
            proxies (dict, optional): Http request proxies. Defaults to {}.
            history_offset (int, optional): Limit conversation history to this number of last texts. Defaults to 10250.
            act (str|int, optional): Awesome prompt key or index. (Used as intro). Defaults to None.
            quiet (bool, optional): Ignore web search-results and yield final response only. Defaults to False.
        """
        self.max_tokens_to_sample = max_tokens
        self.is_conversation = is_conversation
        self.last_response = {}
        self.web_results: dict = {}
        self.quiet = quiet

        self.session: Session = Session()
        self.user_agent: dict = {
            "User-Agent": "Ask/2.9.1/2406 (iOS; iPhone; Version 17.1) isiOSOnMac/false",
            "X-Client-Name": "Perplexity-iOS",
            "X-App-ApiClient": "ios",
        }
        self.session.headers.update(self.user_agent)

        if email and ".perplexity_session" in listdir():
            self._recover_session(email)
        else:
            self._init_session_without_login()

            if email:
                self._login(email)

        self.email: str = email
        self.t: str = self._get_t()
        self.sid: str = self._get_sid()

        self.n: int = 1
        self.base: int = 420
        self.queue: list = []
        self.finished: bool = True
        self.last_uuid: str = None
        self.backend_uuid: str = (
            None  # unused because we can't yet follow-up questions
        )
        self.frontend_session_id: str = str(uuid4())

        assert self._ask_anonymous_user(), "failed to ask anonymous user"
        self.ws: WebSocketApp = self._init_websocket()
        self.ws_thread: Thread = Thread(target=self.ws.run_forever).start()
        self._auth_session()

        while not (self.ws.sock and self.ws.sock.connected):
            sleep(0.01)

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        self.session.proxies = proxies

    def _recover_session(self, email: str) -> None:
        with open(".perplexity_session", "r") as f:
            perplexity_session: dict = loads(f.read())

        if email in perplexity_session:
            self.session.cookies.update(perplexity_session[email])
        else:
            self._login(email, perplexity_session)

    def _login(self, email: str, ps: dict = None) -> None:
        self.session.post(
            url="https://www.perplexity.ai/api/auth/signin-email",
            data={"email": email},
        )

        email_link: str = str(input("paste the link you received by email: "))
        self.session.get(email_link)

        if ps:
            ps[email] = self.session.cookies.get_dict()
        else:
            ps = {email: self.session.cookies.get_dict()}

        with open(".perplexity_session", "w") as f:
            f.write(dumps(ps))

    def _init_session_without_login(self) -> None:
        self.session.get(url=f"https://www.perplexity.ai/search/{str(uuid4())}")
        self.session.headers.update(self.user_agent)

    def _auth_session(self) -> None:
        self.session.get(url="https://www.perplexity.ai/api/auth/session")

    def _get_t(self) -> str:
        return format(getrandbits(32), "08x")

    def _get_sid(self) -> str:
        return loads(
            self.session.get(
                url=f"https://www.perplexity.ai/socket.io/?EIO=4&transport=polling&t={self.t}"
            ).text[1:]
        )["sid"]

    def _ask_anonymous_user(self) -> bool:
        response = self.session.post(
            url=f"https://www.perplexity.ai/socket.io/?EIO=4&transport=polling&t={self.t}&sid={self.sid}",
            data='40{"jwt":"anonymous-ask-user"}',
        ).text

        return response == "OK"

    def _start_interaction(self) -> None:
        self.finished = False

        if self.n == 9:
            self.n = 0
            self.base *= 10
        else:
            self.n += 1

        self.queue = []

    def _get_cookies_str(self) -> str:
        cookies = ""
        for key, value in self.session.cookies.get_dict().items():
            cookies += f"{key}={value}; "
        return cookies[:-2]

    def _write_file_url(self, filename: str, file_url: str) -> None:
        if ".perplexity_files_url" in listdir():
            with open(".perplexity_files_url", "r") as f:
                perplexity_files_url: dict = loads(f.read())
        else:
            perplexity_files_url: dict = {}

        perplexity_files_url[filename] = file_url

        with open(".perplexity_files_url", "w") as f:
            f.write(dumps(perplexity_files_url))

    def _init_websocket(self) -> WebSocketApp:
        def on_open(ws: WebSocketApp) -> None:
            ws.send("2probe")
            ws.send("5")

        def on_message(ws: WebSocketApp, message: str) -> None:
            if message == "2":
                ws.send("3")
            elif not self.finished:
                if message.startswith("42"):
                    message: list = loads(message[2:])
                    content: dict = message[1]
                    if "mode" in content and content["mode"] == "copilot":
                        content["copilot_answer"] = loads(content["text"])
                    elif "mode" in content:
                        content.update(loads(content["text"]))
                    content.pop("text")
                    if (
                        not ("final" in content and content["final"])
                    ) or ("status" in content and content["status"] == "completed"):
                        self.queue.append(content)
                    if message[0] == "query_answered":
                        self.last_uuid = content["uuid"]
                        self.finished = True
                elif message.startswith("43"):
                    message: dict = loads(message[3:])[0]
                    if (
                        "uuid" in message and message["uuid"] != self.last_uuid
                    ) or "uuid" not in message:
                        self.queue.append(message)
                        self.finished = True

        return WebSocketApp(
            url=f"wss://www.perplexity.ai/socket.io/?EIO=4&transport=websocket&sid={self.sid}",
            header=self.user_agent,
            cookie=self._get_cookies_str(),
            on_open=on_open,
            on_message=on_message,
            on_error=lambda ws, err: print(f"websocket error: {err}"),
        )

    def _s(
        self,
        query: str,
        mode: str = "concise",
        search_focus: str = "internet",
        attachments: list[str] = [],
        language: str = "en-GB",
        in_page: str = None,
        in_domain: str = None,
    ) -> None:
        assert self.finished, "already searching"
        assert mode in ["concise", "copilot"], "invalid mode"
        assert len(attachments) <= 4, "too many attachments: max 4"
        assert (
            search_focus
            in [
                "internet",
                "scholar",
                "writing",
                "wolfram",
                "youtube",
                "reddit",
            ]
        ), "invalid search focus"

        if in_page:
            search_focus = "in_page"
        if in_domain:
            search_focus = "in_domain"

        self._start_interaction()
        ws_message: str = (
            f"{self.base + self.n}"
            + dumps(
                [
                    "perplexity_ask",
                    query,
                    {
                        "version": "2.1",
                        "source": "default",  # "ios"
                        "frontend_session_id": self.frontend_session_id,
                        "language": language,
                        "timezone": "CET",
                        "attachments": attachments,
                        "search_focus": search_focus,
                        "frontend_uuid": str(uuid4()),
                        "mode": mode,
                        # "use_inhouse_model": True
                        "in_page": in_page,
                        "in_domain": in_domain,
                    },
                ]
            )
        )

        self.ws.send(ws_message)

    def search(
        self,
        query: str,
        mode: str = "concise",
        search_focus: str = "internet",
        attachments: list[str] = [],
        language: str = "en-GB",
        timeout: float = 30,
        in_page: str = None,
        in_domain: str = None,
    ) -> Iterable[Dict]:
        self._s(query, mode, search_focus, attachments, language, in_page, in_domain)

        start_time: float = time()
        while (not self.finished) or len(self.queue) != 0:
            if timeout and time() - start_time > timeout:
                self.finished = True
                return {"error": "timeout"}
            if len(self.queue) != 0:
                yield self.queue.pop(0)

    def search_sync(
        self,
        query: str,
        mode: str = "concise",
        search_focus: str = "internet",
        attachments: list[str] = [],
        language: str = "en-GB",
        timeout: float = 30,
        in_page: str = None,
        in_domain: str = None,
    ) -> dict:
        self._s(query, mode, search_focus, attachments, language, in_page, in_domain)

        start_time: float = time()
        while not self.finished:
            if timeout and time() - start_time > timeout:
                self.finished = True
                return {"error": "timeout"}

        return self.queue.pop(-1)

    def upload(self, filename: str) -> str:
        assert self.finished, "already searching"
        assert filename.split(".")[-1] in [
            "txt",
            "pdf",
        ], "invalid file format"

        if filename.startswith("http"):
            file = get(filename).content
        else:
            with open(filename, "rb") as f:
                file = f.read()

        self._start_interaction()
        ws_message: str = (
            f"{self.base + self.n}"
            + dumps(
                [
                    "get_upload_url",
                    {
                        "version": "2.1",
                        "source": "default",
                        "content_type": "text/plain"
                        if filename.split(".")[-1] == "txt"
                        else "application/pdf",
                    },
                ]
            )
        )

        self.ws.send(ws_message)

        while not self.finished or len(self.queue) != 0:
            if len(self.queue) != 0:
                upload_data = self.queue.pop(0)

        assert not upload_data["rate_limited"], "rate limited"

        post(
            url=upload_data["url"],
            files={
                "acl": (None, upload_data["fields"]["acl"]),
                "Content-Type": (None, upload_data["fields"]["Content-Type"]),
                "key": (None, upload_data["fields"]["key"]),
                "AWSAccessKeyId": (None, upload_data["fields"]["AWSAccessKeyId"]),
                "x-amz-security-token": (
                    None,
                    upload_data["fields"]["x-amz-security-token"],
                ),
                "policy": (None, upload_data["fields"]["policy"]),
                "signature": (None, upload_data["fields"]["signature"]),
                "file": (filename, file),
            },
        )

        file_url: str = (
            upload_data["url"] + upload_data["fields"]["key"].split("$")[0] + filename
        )

        self._write_file_url(filename, file_url)

        return file_url

    def threads(self, query: str = None, limit: int = None) -> list[dict]:
        assert self.email, "not logged in"
        assert self.finished, "already searching"

        if not limit:
            limit = 20
        data: dict = {"version": "2.1", "source": "default", "limit": limit, "offset": 0}
        if query:
            data["search_term"] = query

        self._start_interaction()
        ws_message: str = f"{self.base + self.n}" + dumps(["list_ask_threads", data])

        self.ws.send(ws_message)

        while not self.finished or len(self.queue) != 0:
            if len(self.queue) != 0:
                return self.queue.pop(0)

    def list_autosuggest(self, query: str = "", search_focus: str = "internet") -> list[dict]:
        assert self.finished, "already searching"

        self._start_interaction()
        ws_message: str = (
            f"{self.base + self.n}"
            + dumps(
                [
                    "list_autosuggest",
                    query,
                    {
                        "has_attachment": False,
                        "search_focus": search_focus,
                        "source": "default",
                        "version": "2.1",
                    },
                ]
            )
        )

        self.ws.send(ws_message)

        while not self.finished or len(self.queue) != 0:
            if len(self.queue) != 0:
                return self.queue.pop(0)

    def close(self) -> None:
        self.ws.close()

        if self.email:
            with open(".perplexity_session", "r") as f:
                perplexity_session: dict = loads(f.read())

            perplexity_session[self.email] = self.session.cookies.get_dict()

            with open(".perplexity_session", "w") as f:
                f.write(dumps(perplexity_session))

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> dict | Generator:
        """Chat with AI

                Args:
                    prompt (str): Prompt to be send.
                    stream (bool, optional): Flag for streaming response. Defaults to False.
                    raw (bool, optional): Stream back raw response as received. Defaults to False.
                    optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
                    conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
                Returns:
                   dict : {}
                ```json
        {
            "status": "pending",
            "uuid": "3604dfcc-611f-4b7d-989d-edca2a7233c7",
            "read_write_token": null,
            "frontend_context_uuid": "f6d43119-5231-481d-b692-f52e1f52d2c6",
            "final": false,
            "backend_uuid": "a6d6ec9e-da69-4841-af74-0de0409267a8",
            "media_items": [],
            "widget_data": [],
            "knowledge_cards": [],
            "expect_search_results": "false",
            "mode": "concise",
            "search_focus": "internet",
            "gpt4": false,
            "display_model": "turbo",
            "attachments": null,
            "answer": "",
            "web_results": [],
            "chunks": [],
            "extra_web_results": []
        }
                ```
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )

        def for_stream():
            for response in self.search(conversation_prompt):
                yield dumps(response) if raw else response
                self.last_response.update(response)

            self.conversation.update_chat_history(
                prompt, self.get_message(self.last_response)
            )

        def for_non_stream():
            self.last_response.update(self.search_sync(conversation_prompt))
            self.conversation.update_chat_history(
                prompt, self.get_message(self.last_response)
            )
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> str | Generator:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
        Returns:
            str: Response generated
        """

        def for_stream():
            for response in self.ask(
                prompt, True, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            return self.get_message(
                self.ask(
                    prompt,
                    False,
                    optimizer=optimizer,
                    conversationally=conversationally,
                )
            )

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: dict) -> str:
        """Retrieves message only from response

        Args:
            response (dict): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        assert isinstance(response, dict), "Response should be of dict data-type only"
        text_str: str = response.get("answer", "")

        def update_web_results(web_results: list) -> None:
            for index, results in enumerate(web_results, start=1):
                self.web_results[str(index) + ". " + results["name"]] = dict(
                    url=results.get("url"), snippet=results.get("snippet")
                )

        if response.get("text"):
            # last chunk
            target: dict[str, Any] = json.loads(response.get("text"))
            text_str = target.get("answer")
            web_results: list[dict] = target.get("web_results")
            self.web_results.clear()
            update_web_results(web_results)

            return text_str 

        else:
            return text_str 


if __name__ == "__main__":
    perplexity = Perplexity() 
    # Stream the response
    response = perplexity.chat("What is the meaning of life?")
    for chunk in response:
        print(chunk, end="", flush=True)

    perplexity.close()